"""Bluetooth Low Energy display helpers.
"""

def print_data_and_hex(data, is_handle_in_data, prefix=""):
    """Print supplied data followed by the hexadecimal equivalent

    :param data: List of data strings to be printed or string
    :param is_handle_in_data: If data supplied contains the source handle sending
                              the data, such as when we read data from a device
                              using a UUID, we need to know to print the handle
                              separate from the handle
    :param prefix: Prefix all printed lines with the supplied string
    :type data: list of strings or string
    :type is_handle_in_data: bool
    :type prefix: str

    """
    print(prefix + "Data (Copy/Paste version)")
    print(prefix + "==========================")
    if data == -1:
        print(prefix + "Invalid Handle/UUID")
        print(prefix + "=====")
        return
    if data == -2:
        print(prefix + "Permission error. Cannot read supplied handle/UUID.")
        print(prefix + "=====")
        return

    if data is None:
        print(prefix + "No data found from a previous from a previous read operation.")
        print(prefix + "=====")
        return
    if isinstance(data, list):
        for i in data:
            if is_handle_in_data:
                #UUID read response packets contain
                #the originating header in the first two bytes
                #handle reverse order when received
                handle = i[:2][::-1]
                i = i[2:]
                if handle is not None:
                    print(prefix + "Handle:", "".join(f"{ord(c):02x}" for c in handle))
                else:
                    print(prefix + "Handle:")
            print(prefix + str(i))
            print(prefix + "-" * len(str(i)))
            print(prefix + " ". join(f"{ord(c):02x}" for c in str(i)))
    else:
        if is_handle_in_data:
            # UUID read response packets contain
            # the originating header in the first two bytes
            # handle reverse order when received
            handle = data[:2][::-1]
            i = data[2:]
            if handle is not None:
                print(prefix + "Handle:", "".join(f"{ord(c):02x}" for c in handle))
            else:
                print(prefix + "Handle:")
        print(prefix + str(data))
        print(prefix + "-" * len(str(data)))
        print(prefix + " ".join(f"{ord(c):02x}" for c in str(data)))
    print(prefix + "=====")



"""
PCAP virtual device for WHAD.

This module provides a PCAP reader that behaves like a virtual WHAD device by
replaying packets from a specific PCAP file like they were being received with
some compatible hardware.
"""
import logging

from os.path import exists
from time import sleep
from struct import unpack

from scapy.layers.bluetooth4LE import BTLE
from scapy.layers.dot15d4 import Dot15d4
from scapy.utils import PcapReader

from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady, WhadDeviceAccessDenied, \
    WhadDeviceDisconnected, WhadDeviceError

from whad.hub.generic.cmdresult import CommandResult
from whad.scapy.layers.phy import Phy_Packet
from whad.hub.dot15d4 import Dot15d4Metadata
from whad.hub.ble import BLEMetadata
from whad.hub.esb import ESBMetadata
from whad.hub.phy import PhyMetadata, Modulation, Endianness
from whad.hub.unifying import UnifyingMetadata
from whad.hub.discovery import Domain
from whad.ble.utils.phy import FieldsSize

from ..device import VirtualDevice
from .capabilities import CAPABILITIES

logger = logging.getLogger(__name__)

class Pcap(VirtualDevice):
    """PCAP replay virtual device implementation.
    """

    INTERFACE_NAME = "pcap"

    @classmethod
    def list(cls):
        '''
        Returns a list of available PCAP devices.
        '''
        return None

    @classmethod
    def check_interface(cls, interface):
        '''
        This method checks dynamically if the provided interface can be instantiated.
        '''
        logger.info("Checking interface: %s", str(interface))
        return interface.endswith(".pcap") or interface.endswith(".pcapng")

    @property
    def identifier(self):
        '''
        Returns the identifier of the current device (e.g., bus + address in
        format "<bus>-<address>").
        '''
        return "pcap:" + str(self.__filename)


    def __init__(self, filename):
        """
        Create device connection
        """
        self.__opened = False
        self.__started = False
        self.__flush = "flush:" in filename
        self.__filename = filename.replace("flush:", "")
        self.__pcap_reader = None
        self.__dlt = None
        self.__domain = None
        self.__start_timestamp, self.__last_timestamp = None, None


        self.__supported_frequency_range = [
            (281000000, 361000000),
            (378000000, 481000000),
            (749000000, 962000000),
            (2400000000, 2500000000)
        ]
        super().__init__()

    def _is_reader(self):
        """
        Returns True if the PCAP is in reading mode.
        """
        return self.__pcap_reader is not None

    def _get_dlt(self):
        if self._is_reader():
            if hasattr(self.__pcap_reader, "linktype"):
                dlt = self.__pcap_reader.linktype
            else:
                # PCAP-ng
                _ = self.__pcap_reader.read_packet()
                dlt = self.__pcap_reader.interfaces[0][0]
                self.__pcap_reader = PcapReader(self.__filename)
            return dlt

        # No DLT
        return None

    def _get_domain(self):
        """Retrieve the domain associated with the current link-layer type (DLT).
        """
        if self.__dlt in CAPABILITIES:
            return list(CAPABILITIES[self.__dlt][0].keys())[0]
        else:
            return None

    def open(self):
        try:
            if exists(self.__filename):
                logger.info("Existing PCAP file")
                self.__pcap_reader = PcapReader(self.__filename)
                self.__dlt = self._get_dlt()
                self.__domain = self._get_domain()
                if self.__domain is None:
                    raise WhadDeviceError(f"Unsupported PCAP file (DLT: {self.__dlt})")
            else:
                logger.info("No PCAP file")
                raise WhadDeviceNotFound("pcap")
        except WhadDeviceError as deverr:
            raise deverr
        except Exception as access_error:
            raise WhadDeviceAccessDenied("pcap") from access_error

        self._dev_id = self._get_serial_number()
        self._fw_author = self._get_manufacturer()
        self._fw_url = self._get_url()
        self._fw_version = self._get_firmware_version()
        self._dev_capabilities = self._get_capabilities()
        #self.__flush = False
        self.__opened = True
        # Ask parent class to run a background I/O thread
        super().open()

    def write(self, payload):
        if not self.__opened:
            raise WhadDeviceNotReady()

    def read(self) -> bytes:
        """Read packets from PCAP file and replay them
        """
        if not self.__opened:
            logger.error("Device not ready")
            raise WhadDeviceNotReady()
        while self.__started:
            try:
                if self._is_reader():
                    pkt = self.__pcap_reader.read_packet()
                    return self.__to_raw_message(pkt)
            except EOFError as eof:
                # End of PCAP reached, we are no more open and notify the
                # calling thread that we are now disconnected.
                logger.debug("[PCAPDevice] EOF reached")
                self.__opened = False
                raise WhadDeviceDisconnected() from eof

    @property
    def opened(self) -> bool:
        return self.__opened

    def reset(self):
        pass

    def _generate_metadata(self, pkt):
        if self.__domain == Domain.Dot15d4:
            metadata = Dot15d4Metadata.convert_from_header(pkt)
        elif self.__domain == Domain.BtLE:
            metadata = BLEMetadata.convert_from_header(pkt)
        elif self.__domain == Domain.Esb:
            metadata = ESBMetadata.convert_from_header(pkt)
        elif self.__domain == Domain.LogitechUnifying:
            metadata = UnifyingMetadata.convert_from_header(pkt)
        elif self.__domain == Domain.Phy:
            metadata = PhyMetadata.convert_from_header(pkt)

        else:
            return None
        if self.__start_timestamp is None:
            self.__start_timestamp = metadata.timestamp
        metadata.timestamp = metadata.timestamp - self.__start_timestamp
        return metadata

    def _interframe_delay(self, timestamp):
        if not self.__flush:
            if self.__last_timestamp is None:
                self.__last_timestamp = 0
            sleep((timestamp - self.__last_timestamp)/100000)

    def __to_raw_message(self, pkt) -> bytes:
        msg = None
        if self.__domain == Domain.Dot15d4:
            metadata = self._generate_metadata(pkt)
            self._interframe_delay(metadata.timestamp)
            self.__last_timestamp = metadata.timestamp
            msg = self.__to_whad_zigbee_raw_pdu(bytes(pkt[Dot15d4]), channel=metadata.channel,
                                           lqi=metadata.lqi, rssi=metadata.rssi,
                                           timestamp=metadata.timestamp)
        elif self.__domain == Domain.BtLE:
            metadata = self._generate_metadata(pkt)
            self._interframe_delay(metadata.timestamp)
            self.__last_timestamp = metadata.timestamp
            msg = self.__to_whad_ble_raw_pdu(pkt, metadata)

        elif self.__domain == Domain.Esb:
            metadata = self._generate_metadata(pkt)
            self._interframe_delay(metadata.timestamp)
            self.__last_timestamp = metadata.timestamp
            msg = self.__to_whad_esb_raw_pdu(pkt, metadata)

        elif self.__domain == Domain.LogitechUnifying:
            metadata = self._generate_metadata(pkt)
            self._interframe_delay(metadata.timestamp)
            self.__last_timestamp = metadata.timestamp
            msg = self.__to_whad_unifying_raw_pdu(pkt, metadata)

        elif self.__domain == Domain.Phy:
            metadata = self._generate_metadata(pkt)
            self._interframe_delay(metadata.timestamp)
            self.__last_timestamp = metadata.timestamp
            msg = self.__to_whad_phy_pdu(pkt, metadata)

        if msg is not None:
            # Serialize protobuf message
            raw_msg = msg.serialize()

            # Define header
            header = [
                0xAC, 0xBE,
                len(raw_msg) & 0xff,
                (len(raw_msg) >> 8) & 0xff
            ]

            # Build the final payload
            return bytes(header) + raw_msg
        else:
            return b""

    def __to_whad_phy_pdu(self, packet, metadata):
        return self.hub.phy.create_packet_received(
            metadata.frequency, # TODO: frequency,
            bytes(packet[Phy_Packet]),
            metadata.rssi, # TODO: rssi
            metadata.timestamp,
            metadata.syncword,
            metadata.datarate,
            metadata.deviation,
            Modulation(metadata.modulation),
            Endianness(metadata.endianness)
        )

    def __to_whad_unifying_raw_pdu(self, packet, metadata):
        # Create a RawPduReceived message
        return self.hub.unifying.create_raw_pdu_received(
            metadata.channel,
            bytes(packet),
            metadata.rssi,
            metadata.timestamp,
            metadata.is_crc_valid,
            metadata.address
        )

    def __to_whad_esb_raw_pdu(self, packet, metadata):

        # Create a RawPduReceived message
        return self.hub.esb.create_raw_pdu_received(
            metadata.channel,
            bytes(packet),
            metadata.rssi,
            metadata.timestamp,
            metadata.is_crc_valid,
            metadata.address
        )

    def __to_whad_ble_raw_pdu(self, packet, metadata):
        packet = packet[BTLE:]
        access_address = packet.access_addr
        pdu = bytes(packet)[FieldsSize.ACCESS_ADDRESS_SIZE:-FieldsSize.CRC_SIZE]

        # Create a RawPduReceived message
        return self.hub.ble.create_raw_pdu_received(
            metadata.direction,
            pdu,
            access_address,
            0,
            crc_validity=metadata.is_crc_valid,
            crc=packet.crc,
            channel=metadata.channel,
            timestamp=metadata.timestamp,
            rssi=metadata.rssi
        )

    # Virtual device whad message builder
    def __to_whad_zigbee_raw_pdu(self, packet, channel=None, rssi=None, lqi=None,
                                  is_fcs_valid=True, timestamp=None):
        pdu = packet[:-2]
        fcs = unpack("<H",packet[-2:])[0]

        # Create a RawPduReceived message
        msg = self.hub.dot15d4.create_raw_pdu_received(
            channel,
            pdu,
            fcs,
            lqi = lqi,
            fcs_validity=is_fcs_valid
        )

        # Set timestamp and RSSI if provided
        if rssi is not None:
            msg.rssi = rssi
        if timestamp is not None:
            msg.timestamp = timestamp

        return msg


    # Virtual device whad message callbacks
    def _on_whad_ble_stop(self, message): # pylint: disable=W0613
        self.__started = False
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_ble_start(self, message): # pylint: disable=W0613
        self.__started = True
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_ble_sniff_adv(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_ble_sniff_conn(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)


    def _on_whad_phy_stop(self, message): # pylint: disable=W0613
        self.__started = False
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_send_raw(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_sniff(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_start(self, message): # pylint: disable=W0613
        self.__started = True
        self._send_whad_command_result(CommandResult.SUCCESS)


    def _on_whad_phy_get_supported_freq(self, message): # pylint: disable=W0613
        # Create a SupportedFreqRanges
        msg = self.hub.phy.create_supported_freq_ranges(
            self.__supported_frequency_range
        )

        # Send message
        self._send_whad_message(msg)

    def _on_whad_phy_send(self, message):
        self._send_packet(message.packet)
        self._send_whad_command_result(CommandResult.SUCCESS)


    def _on_whad_phy_set_freq(self, message):
        found = False
        for supported_range in self.__supported_frequency_range:
            range_start, range_end = supported_range[0], supported_range[1]
            if message.frequency >= range_start and message.frequency <= range_end:
                found = True
                break
        if found:
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.PARAMETER_ERROR)

    def _on_whad_phy_datarate(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_packet_size(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_endianness(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_tx_power(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_mod_ask(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_mod_4fsk(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_mod_fsk(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_mod_gfsk(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_phy_sync_word(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)


    def _on_whad_dot15d4_stop(self, message): # pylint: disable=W0613
        self.__started = False
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_dot15d4_send_raw(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_dot15d4_sniff(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_dot15d4_start(self, message): # pylint: disable=W0613
        self.__started = True
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_stop(self, message): # pylint: disable=W0613
        self.__started = False
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_send_raw(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_sniff(self, message): # pylint: disable=W0613
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_esb_start(self, message): # pylint: disable=W0613
        self.__started = True
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_unifying_stop(self, message): # pylint: disable=W0613
        self.__domain = Domain.LogitechUnifying
        self.__started = False
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_unifying_send_raw(self, message): # pylint: disable=W0613
        self.__domain = Domain.LogitechUnifying
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_unifying_sniff(self, message): # pylint: disable=W0613
        self.__domain = Domain.LogitechUnifying
        self._send_whad_command_result(CommandResult.SUCCESS)

    def _on_whad_unifying_start(self, message): # pylint: disable=W0613
        self.__domain = Domain.LogitechUnifying
        self.__started = True
        self._send_whad_command_result(CommandResult.SUCCESS)

    # Discovery related functions
    def _get_capabilities(self):
        index = 0 if self._is_reader() else 1
        capabilities = CAPABILITIES[self.__dlt][index]
        return capabilities

    def _get_manufacturer(self):
        return "whad-team".encode('utf-8')

    def _get_serial_number(self):
        return bytes.fromhex("00" * 16)

    def _get_firmware_version(self):
        return (0, 0, 0)

    def _get_url(self):
        return "https://github.com/whad-team/whad-client".encode('utf-8')



"""BLE peripheral emulation interactive shell
"""
import json

from whad.cli.app import command
from whad.ble.cli.peripheral.shell import BlePeriphShell

def check_profile(profile):
    """Check profile validity.

    :param dict p: Profile
    :return bool: True if profile format is OK, False otherwise
    """
    # Make sure we have a 'devinfo' entry
    if 'devinfo' not in profile:
        return False

    # Make sure we have a valid adv_data entry
    if 'adv_data' not in profile['devinfo']:
        return False
    try:
        bytes.fromhex(profile["devinfo"]["adv_data"])
    except ValueError:
        return False

    # Make sure we have a valid scan_rsp entry (if provided)
    if 'scan_rsp' not in profile['devinfo']:
        return False
    if profile['devinfo']['scan_rsp'] is not None:
        try:
            bytes.fromhex(profile["devinfo"]["scan_rsp"])
        except ValueError:
            return False

    # Make sure we have a BD address
    if 'bd_addr' not in profile['devinfo']:
        return False

    # OK
    return True

@command('interactive')
def interactive_handler(app, _):
    """interactive BLE shell

    <ansicyan><b>interactive</b></ansicyan>

    Starts an interactive shell and let you create and advertise a BLE WHAD device:
    - create, list and remove services
    - create, list and remove characteristics
    - set characteristic value, send notifications/indications
    """
    # We need to have an interface specified
    if app.interface is not None:
        # If a profile has been provided, load it
        if app.args.profile is not None:
            # Read profile
            with open(app.args.profile,'rb') as f:
                profile_json = f.read()
            try:
                if not check_profile(json.loads(profile_json)):
                    app.error("Invalid JSON file (does not contain a valid GATT profile).")
                    profile_json = None
            except json.decoder.JSONDecodeError as parsing_err:
                app.error((f"Invalid JSON file, parsing error line {parsing_err.lineno}: "
                          f"{parsing_err.msg}"))
                app.exit()
        else:
            profile_json = None

        # Launch an interactive shell
        myshell = BlePeriphShell(app.interface, profile_json)
        myshell.run()
    else:
        app.error("You need to specify an interface with option --interface.")



"""Helpers for BLE GATT central CLI tool.
"""
import json
from typing import Tuple
from argparse import Namespace
from prompt_toolkit import print_formatted_text, HTML
from whad.hub.ble.bdaddr import BDAddress
from whad.ble import Central
from whad.ble.stack.att.exceptions import AttError, AttributeNotFoundError, \
    InsufficientAuthenticationError, InsufficientAuthorizationError, \
    InsufficientEncryptionKeySize, ReadNotPermittedError, \
    WriteNotPermittedError, InvalidHandleValueError

def show_att_error(app, error: AttError):
    """Parse ATT error and show exception.
    """
    if isinstance(error, InvalidHandleValueError):
        app.error('ATT Error: wrong value handle')
    elif isinstance(error, ReadNotPermittedError):
        app.error('ATT error: read operation not allowed')
    elif isinstance(error, WriteNotPermittedError):
        app.error('ATT error: write operation not allowed')
    elif isinstance(error, InsufficientAuthenticationError):
        app.error('ATT error: insufficient authentication')
    elif isinstance(error, InsufficientAuthorizationError):
        app.error('ATT error: insufficient authorization')
    elif isinstance(error, AttributeNotFoundError):
        app.error('ATT error: attribute not found')
    elif isinstance(error, InsufficientEncryptionKeySize):
        app.error('ATT error: insufficient encryption')


def set_bd_address(app, central: Central):
    """Set central BLE address
    """
    bd_addr = None
    is_public = False

    # If a spoofed address has been provided, then try to set it
    if app.args.bdaddr_pub_src is not None:
        # Make sure it is a valid BD address
        if BDAddress.check(app.args.bdaddr_pub_src):
            bd_addr = app.args.bdaddr_pub_src
            is_public = True
    elif app.args.bdaddr_rand_src is not None:
        # Make sure it is a valid BD address
        if BDAddress.check(app.args.bdaddr_rand_src):
            bd_addr = app.args.bdaddr_rand_src
            is_public = False

    # Set the BD address
    if central.set_bd_address(bd_addr, public=is_public):
        print_formatted_text(HTML(
            f"BLE source address set to <b>{app.args.bdaddr_src.lower()}</b>"
        ))
    else:
        app.warning(("Cannot spoof BD address, please make sure your WHAD "
                     "interface supports this feature."))


def create_central(app, piped: bool = False) -> Tuple[Central, dict]:
    """Create central connector.
    """
    central = None
    profile_loaded = False

    # Is app stdin piped ?
    if piped:
        # Create connection structure
        initiator = BDAddress(
            str(app.args.initiator_bdaddr),
            addr_type=int(app.args.initiator_addrtype)
        )
        advertiser = BDAddress(
            str(app.args.initiator_bdaddr),
            addr_type=int(app.args.initiator_addrtype)
        )
        existing_connection = Namespace(
            initiator=initiator.value,
            init_addr_type=int(app.args.initiator_addrtype),
            advertiser=advertiser.value,
            adv_addr_type=int(app.args.target_addrtype),
            conn_handle=int(app.args.conn_handle)
        )

        # Create central and populate GATT profile if required
        if app.args.profile is not None:
            # Load profile
            try:
                # Load file content
                with open(app.args.profile,'rb') as profile:
                    profile_json = profile.read()

                # Create central with GATT profile information and current
                # connection information
                central = Central(app.input_interface, existing_connection, from_json=profile_json)

                # Profile has been successfully loaded from JSON
                profile_loaded = True
            except IOError:
                app.error(f"Cannot access profile file ({app.args.profile})")
            except json.decoder.JSONDecodeError:
                app.error(f"Cannot parse profile file ({app.args.profile})")
        else:
            # No GATT profile, create a classic Central connector with current
            # connection information
            central = Central(app.input_interface, existing_connection)
    else:
        if app.args.profile is not None:
            # Load profile
            try:
                # Load file content
                with open(app.args.profile,'rb') as profile:
                    profile_json = profile.read()

                # Create Central connector with provided GATT profile
                central = Central(app.interface, from_json=profile_json)

                # Set source BD address if required
                if app.args.bdaddr_pub_src is not None or app.args.bdaddr_rand_src is not None:
                    set_bd_address(app, central)

                # Profile has been successfully loaded from JSON
                profile_loaded = True
            except IOError:
                app.error(f"Cannot access profile file ({app.args.profile})")
            except json.decoder.JSONDecodeError:
                app.error(f"Cannot parse profile file ({app.args.profile})")
        else:
            # Create classic Central connector
            central = Central(app.interface)

            # Set source BD address if required
            if app.args.bdaddr_pub_src is not None or app.args.bdaddr_rand_src is not None:
                set_bd_address(app, central)


    return (central, profile_loaded)



from scapy.fields import (
    Field, XByteField, ShortField, ByteField, LEShortField, LEIntField,
)
from scapy.config import conf
from scapy.packet import Packet, bind_layers
from scapy.utils import str2mac, mac2str
from scapy.volatile import RandMAC
from scapy.layers.bluetooth4LE import BTLE

class LEMACField(Field):
    def __init__(self, name, default):
        Field.__init__(self, name, default, "6s")
    def i2m(self, pkt, x):
        if x is None:
            return "\0\0\0\0\0\0"
        return mac2str(x)[::-1]
    def m2i(self, pkt, x):
        return str2mac(x[::-1])
    def any2i(self, pkt, x):
        if type(x) is str and len(x) == 6:
            x = self.m2i(pkt, x)
        return x
    def i2repr(self, pkt, x):
        x = self.i2h(pkt, x)
        if self in conf.resolve:
            x = conf.manufdb._resolve_MAC(x)
        return x
    def randval(self):
        return RandMAC()

class NordicBLE(Packet):
    name="Nordic BLE"
    fields_desc = [
            XByteField('board_id',-1),
            ByteField('proto_version', -1),
            LEShortField('pkt_counter', -1),
            LEShortField('pkt_id', -1),
            ShortField('pkt_len',-1),
            XByteField('flags', 0),
            ByteField('channel', 0),
            ByteField('rssi', 0),
            ShortField('event_counter', 0),
            LEIntField('timestamp',0),
        ]

    def mysummary(self):
        return self.sprintf("NordicBLE channel=%channel%")

bind_layers( NordicBLE, BTLE,)
conf.l2types.register(272, NordicBLE)



from whad.zigbee.stack.nwk.constants import ZigbeeDeviceType
from whad.zigbee.profile.nodes import CoordinatorNode, EndDeviceNode, RouterNode
from whad.dot15d4.address import Dot15d4Address
from random import randint
from time import sleep
import logging

logger = logging.getLogger(__name__)

class JoiningForbidden(Exception):
    """
    Exception raised when joining is forbidden on the network.
    """
    pass

class NotAssociated(Exception):
    """
    Exception raised when we are not associated in this network.
    """
    pass

class NotAuthorized(Exception):
    """
    Exception raised when we are not authorized in this network.
    """
    pass

class Network:
    """
    Class representing a ZigBee network.
    """
    def __init__(self, nwkObject=None, stack = None):
        # reference to the corresponding network representation
        # in the stack, we are defining an helper class here.
        self.__nwk_object = nwkObject

        # reference to the stack
        self.__stack = stack

        # only one coordinator, but possibly multiple routers & end devices
        self.coordinator = None
        self.routers = []
        self.end_devices = []


    @property
    def nodes(self):
        """
        Returns all the nodes present in this network.
        """
        return (
            [self.coordinator] if self.coordinator is not None else []
        ) + self.routers + self.end_devices

    @property
    def stack(self):
        """
        Returns the associated instance of the stack.
        """
        return self.__stack

    @property
    def pan_id(self):
        """
        Returns the network Pan ID.
        """
        return self.__nwk_object.dot15d4_pan_network.coord_pan_id

    @property
    def extended_pan_id(self):
        """
        Returns the network extended Pan ID.
        """
        return self.__nwk_object.extended_pan_id

    @property
    def channel(self):
        """
        Returns the channel used by the network.
        """
        return self.__nwk_object.channel

    def discover(self):
        """
        Discover all the nodes present in the network.
        """
        if not self.is_associated():
            raise NotAssociated
        if not self.is_authorized():
            raise NotAuthorized
        devices = self.stack.get_layer('apl').get_application_by_name("zdo").device_and_service_discovery.discover_nodes()
        return self.nodes

    def join(self, force: bool = False):
        """
        Join the network (if permitted).

        This method is blocking and wait until we are both associated AND authorized on the network.

        :param force: If set to `True`, join network even if association is not allowed.
        :type force: bool
        """
        if self.is_joining_permitted() or force:
            logger.debug("Start joining network (force=%s)", force)
            join_success = self.stack.get_layer('apl').get_application_by_name("zdo").network_manager.join(self, force=force)
            if join_success:
                while not self.is_authorized():
                    sleep(0.1)
                logger.debug("Successfully joined network")
                return True
            
            logger.debug("Failed joining network")
            return False
        else:
            raise JoiningForbidden

    def rejoin(self, address=None):
        """
        Rejoin the network (if permitted).

        This method is blocking and wait until we are both associated AND authorized on the network.
        """
        if address is None:
            address = randint(0x0001, 0xFFF0)
        print("rejoining...")
        self.stack.get_layer('apl').get_application_by_name("zdo").network_manager.configure_short_address(address)
        rejoin_success = self.stack.get_layer('apl').get_application_by_name("zdo").network_manager.rejoin(self)

        if rejoin_success:
            while not self.is_authorized():
                sleep(0.1)
            return True
        return False

    def is_joining_permitted(self):
        """
        Indicates if joining the network is permitted.
        """
        return self.__nwk_object.joining_permit

    def is_associated(self):
        """
        Indicates if we are associated with the network.
        """
        return self.stack.get_layer('nwk').database.get("nwkExtendedPANID") == self.extended_pan_id

    def is_authorized(self):
        """
        Indicates if we are authorized on the network.
        """
        return self.stack.get_layer('apl').get_application_by_name("zdo").network_manager.authorized

    @property
    def network_key(self):
        """
        Returns the network key associated with this network (if any).
        """
        nwk_material = self.stack.get_layer('nwk').database.get("nwkSecurityMaterialSet")
        if len(nwk_material) > 0:
            return nwk_material[0].key
        else:
            return None

    @network_key.setter
    def network_key(self, value):
        """
        Configure the network key associated with this network.
        """
        self.stack.get_layer('apl').get_application_by_name("zdo").security_manager.provision_network_key(value)

    def leave(self):
        """
        Initiates a leave operation on this network.
        """
        return self.stack.get_layer('apl').get_application_by_name("zdo").network_manager.leave()

    def __eq__(self, other):
        return self.extended_pan_id == other.extended_pan_id

    def __repr__(self):
        return (
            "Network(" +
            "pan_id=" + hex(self.pan_id) + ", " +
            "extended_pan_id=" + str(self.extended_pan_id) + ", " +
            "channel="+str(self.channel) + ", " +
            "joining="+ ("allowed" if self.is_joining_permitted() else "forbidden") + ", " +
            "associated="+ ("yes" if self.is_associated() else "no") + ", "
            "authorized="+ ("yes" if self.is_authorized() else "no") +

            ")"
        )



"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
import os
import logging
from time import sleep

from scapy.config import conf
from scapy.themes import BrightTheme

from whad.cli.ui import wait, success, error
from whad.cli.app import CommandLineApp, run_app
from whad.device import Bridge
from whad.hub import ProtocolHub
from whad.device.unix import UnixConnector, UnixSocketServer
from whad.common.monitors import PcapWriterMonitor

logger = logging.getLogger(__name__)

class WhadDumpApp(CommandLineApp):
    """
    Main `wdump` CLI application class.
    """

    connector = None
    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD PCAP export',
            interface=True,
            commands=False,
            input=CommandLineApp.INPUT_WHAD,
            output=CommandLineApp.OUTPUT_WHAD
        )

        self.add_argument(
            '-f',
            '--force',
            action='store_true',
            dest='force',
            default=False,
            help="Force file overwrite"
        )

        self.add_argument(
            '-a',
            '--append',
            dest='append',
            action='store_true',
            default=False,
            help="Append packets to existing file"
        )

        self.add_argument(
            'pcap',
            help='Pcap file to export'
        )

        # Initialize our monitor.
        self.monitor = None


    def run(self):
        # Launch pre-run tasks
        self.monitor = None
        self.pre_run()

        try:
            if self.args.pcap is not None:

                # If pcap file already exists, ask the user if he/she wants to
                # overwrite it if no '--append' option set
                if os.path.exists(self.args.pcap):
                    if os.path.isfile(self.args.pcap):
                        if not self.args.append:
                            if not self.args.force:
                                self.warning((
                                    "PCAP file already exists, use --append to add packets"
                                    " to it or --force to force overwriting"))
                                return

                            try:
                                # Remove file
                                os.unlink(self.args.pcap)
                            except IOError:
                                self.error("Cannot create PCAP file")
                    else:
                        self.error("Cannot write to PCAP file (not a regular file)")
                        return

                if self.is_piped_interface():
                    interface = self.input_interface
                else:
                    interface = self.interface

                if interface is not None:
                    if not self.args.nocolor:
                        conf.color_theme = BrightTheme()

                    connector = UnixConnector(interface)
                    connector.domain = self.args.domain
                    hub = ProtocolHub(2)
                    connector.format = hub.get(self.args.domain).format

                    self.connector = connector
                    self.monitor = PcapWriterMonitor(self.args.pcap)
                    self.monitor.attach(connector)
                    self.monitor.start()

                    if self.is_stdout_piped():
                        unix_server = UnixConnector(UnixSocketServer(
                            parameters=self.args.__dict__
                        ))

                        while not unix_server.device.opened:
                            if unix_server.device.timedout:
                                return
                            sleep(0.1)

                        # Create our packet bridge
                        logger.info("[wdump] Starting our output pipe")
                        bridge = Bridge(connector, unix_server)
                        bridge.join()
                    else:
                        # Unlock Unix connector first
                        connector.unlock()

                        # Process packets
                        while interface.opened:
                            wait(f"Dumping {self.monitor.packets_written} packets into pcap file: ",
                                suffix = self.args.pcap
                            )
                            sleep(.5)
            else:
                error("You must provide a pcap file.")
        except KeyboardInterrupt:
            # Launch post-run tasks
            if self.monitor is not None:
                self.monitor.stop()
                self.monitor.close()
            self.post_run()

    def post_run(self):
        if not self.is_stdout_piped():
            wait(
                f"Dumping {self.monitor.packets_written} packets into pcap file: ",
                suffix = self.args.pcap,
                end = True
            )
            success(
                f"{self.monitor.packets_written} packets have been dumped into {self.args.pcap}"
            )
        super().post_run()

def wdump_main():
    """Launcher for `wdump` CLI application.
    """
    app = WhadDumpApp()
    run_app(app)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: whad/protocol/dot15d4/dot15d4.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#whad/protocol/dot15d4/dot15d4.proto\x12\x07\x64ot15d4\"P\n\x11SetNodeAddressCmd\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x04\x12*\n\x0c\x61\x64\x64ress_type\x18\x02 \x01(\x0e\x32\x14.dot15d4.AddressType\"\x1b\n\x08SniffCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"%\n\x12\x45nergyDetectionCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\x19\n\x06JamCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\'\n\x07SendCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x0b\n\x03pdu\x18\x02 \x01(\x0c\"7\n\nSendRawCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x0b\n\x03pdu\x18\x02 \x01(\x0c\x12\x0b\n\x03\x66\x63s\x18\x03 \x01(\r\"\x1f\n\x0c\x45ndDeviceCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\x1c\n\tRouterCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"!\n\x0e\x43oordinatorCmd\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\"\n\n\x08StartCmd\"\t\n\x07StopCmd\";\n\x11ManInTheMiddleCmd\x12&\n\x04role\x18\x01 \x01(\x0e\x32\x18.dot15d4.Dot15d4MitmRole\"\x1b\n\x06Jammed\x12\x11\n\ttimestamp\x18\x01 \x01(\x04\":\n\x15\x45nergyDetectionSample\x12\x0e\n\x06sample\x18\x01 \x01(\r\x12\x11\n\ttimestamp\x18\x02 \x01(\x04\"\xc3\x01\n\x0eRawPduReceived\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x19\n\x0c\x66\x63s_validity\x18\x04 \x01(\x08H\x02\x88\x01\x01\x12\x0b\n\x03pdu\x18\x05 \x01(\x0c\x12\x0b\n\x03\x66\x63s\x18\x06 \x01(\r\x12\x10\n\x03lqi\x18\x07 \x01(\rH\x03\x88\x01\x01\x42\x07\n\x05_rssiB\x0c\n\n_timestampB\x0f\n\r_fcs_validityB\x06\n\x04_lqi\"\xb3\x01\n\x0bPduReceived\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\r\x12\x11\n\x04rssi\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x16\n\ttimestamp\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x19\n\x0c\x66\x63s_validity\x18\x04 \x01(\x08H\x02\x88\x01\x01\x12\x0b\n\x03pdu\x18\x05 \x01(\x0c\x12\x10\n\x03lqi\x18\x06 \x01(\rH\x03\x88\x01\x01\x42\x07\n\x05_rssiB\x0c\n\n_timestampB\x0f\n\r_fcs_validityB\x06\n\x04_lqi\"\x9d\x05\n\x07Message\x12\x33\n\rset_node_addr\x18\x01 \x01(\x0b\x32\x1a.dot15d4.SetNodeAddressCmdH\x00\x12\"\n\x05sniff\x18\x02 \x01(\x0b\x32\x11.dot15d4.SniffCmdH\x00\x12\x1e\n\x03jam\x18\x03 \x01(\x0b\x32\x0f.dot15d4.JamCmdH\x00\x12)\n\x02\x65\x64\x18\x04 \x01(\x0b\x32\x1b.dot15d4.EnergyDetectionCmdH\x00\x12 \n\x04send\x18\x05 \x01(\x0b\x32\x10.dot15d4.SendCmdH\x00\x12\'\n\x08send_raw\x18\x06 \x01(\x0b\x32\x13.dot15d4.SendRawCmdH\x00\x12+\n\nend_device\x18\x07 \x01(\x0b\x32\x15.dot15d4.EndDeviceCmdH\x00\x12$\n\x06router\x18\x08 \x01(\x0b\x32\x12.dot15d4.RouterCmdH\x00\x12.\n\x0b\x63oordinator\x18\t \x01(\x0b\x32\x17.dot15d4.CoordinatorCmdH\x00\x12\"\n\x05start\x18\n \x01(\x0b\x32\x11.dot15d4.StartCmdH\x00\x12 \n\x04stop\x18\x0b \x01(\x0b\x32\x10.dot15d4.StopCmdH\x00\x12*\n\x04mitm\x18\x0c \x01(\x0b\x32\x1a.dot15d4.ManInTheMiddleCmdH\x00\x12!\n\x06jammed\x18\r \x01(\x0b\x32\x0f.dot15d4.JammedH\x00\x12\x33\n\ted_sample\x18\x0e \x01(\x0b\x32\x1e.dot15d4.EnergyDetectionSampleH\x00\x12*\n\x07raw_pdu\x18\x0f \x01(\x0b\x32\x17.dot15d4.RawPduReceivedH\x00\x12#\n\x03pdu\x18\x10 \x01(\x0b\x32\x14.dot15d4.PduReceivedH\x00\x42\x05\n\x03msg*\xc5\x01\n\x0e\x44ot15d4Command\x12\x12\n\x0eSetNodeAddress\x10\x00\x12\t\n\x05Sniff\x10\x01\x12\x07\n\x03Jam\x10\x02\x12\x13\n\x0f\x45nergyDetection\x10\x03\x12\x08\n\x04Send\x10\x04\x12\x0b\n\x07SendRaw\x10\x05\x12\x11\n\rEndDeviceMode\x10\x06\x12\x13\n\x0f\x43oordinatorMode\x10\x07\x12\x0e\n\nRouterMode\x10\x08\x12\t\n\x05Start\x10\t\x12\x08\n\x04Stop\x10\n\x12\x12\n\x0eManInTheMiddle\x10\x0b*5\n\x0f\x44ot15d4MitmRole\x12\x13\n\x0fREACTIVE_JAMMER\x10\x00\x12\r\n\tCORRECTOR\x10\x01*&\n\x0b\x41\x64\x64ressType\x12\t\n\x05SHORT\x10\x00\x12\x0c\n\x08\x45XTENDED\x10\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'whad.protocol.dot15d4.dot15d4_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DOT15D4COMMAND._serialized_start=1647
  _DOT15D4COMMAND._serialized_end=1844
  _DOT15D4MITMROLE._serialized_start=1846
  _DOT15D4MITMROLE._serialized_end=1899
  _ADDRESSTYPE._serialized_start=1901
  _ADDRESSTYPE._serialized_end=1939
  _SETNODEADDRESSCMD._serialized_start=48
  _SETNODEADDRESSCMD._serialized_end=128
  _SNIFFCMD._serialized_start=130
  _SNIFFCMD._serialized_end=157
  _ENERGYDETECTIONCMD._serialized_start=159
  _ENERGYDETECTIONCMD._serialized_end=196
  _JAMCMD._serialized_start=198
  _JAMCMD._serialized_end=223
  _SENDCMD._serialized_start=225
  _SENDCMD._serialized_end=264
  _SENDRAWCMD._serialized_start=266
  _SENDRAWCMD._serialized_end=321
  _ENDDEVICECMD._serialized_start=323
  _ENDDEVICECMD._serialized_end=354
  _ROUTERCMD._serialized_start=356
  _ROUTERCMD._serialized_end=384
  _COORDINATORCMD._serialized_start=386
  _COORDINATORCMD._serialized_end=419
  _STARTCMD._serialized_start=421
  _STARTCMD._serialized_end=431
  _STOPCMD._serialized_start=433
  _STOPCMD._serialized_end=442
  _MANINTHEMIDDLECMD._serialized_start=444
  _MANINTHEMIDDLECMD._serialized_end=503
  _JAMMED._serialized_start=505
  _JAMMED._serialized_end=532
  _ENERGYDETECTIONSAMPLE._serialized_start=534
  _ENERGYDETECTIONSAMPLE._serialized_end=592
  _RAWPDURECEIVED._serialized_start=595
  _RAWPDURECEIVED._serialized_end=790
  _PDURECEIVED._serialized_start=793
  _PDURECEIVED._serialized_end=972
  _MESSAGE._serialized_start=975
  _MESSAGE._serialized_end=1644
# @@protoc_insertion_point(module_scope)



"""This file contains the current version number of WHAD.

ATTENTION: This version number must match the one in pyproject.toml !
"""

# Current version number
VERSION_MAJOR = 1
VERSION_MINOR = 2
VERSION_REVISION = 15

def get_version() -> str:
    """Return the current version of this package (WHAD).

    @retval         current version string
    @return-type    str
    """
    return f"{VERSION_MAJOR:d}.{VERSION_MINOR:d}.{VERSION_REVISION:d}"



"""WHAD replay helpers and events.
"""
from whad.exceptions import ReplayInvalidRole

class ReplayRole:
    """Replay role constants
    """
    EMITTER = 1
    RECEIVER = 2
    BOTH = 3

class ReplayInterface:
    """This class provides a template that can be inherited in order to
    implement a replay class.
    """

    def __init__(self, role : int):
        """Initialize a replay interface.

        :param role: Replay role (emitter, receiver or both)
        :type role: int
        """
        # Not connected by default
        self.__connected = False

        # Save replay role
        if role in [ReplayRole.EMITTER, ReplayRole.RECEIVER, ReplayRole.BOTH]:
            self.__role = role
        else:
            raise ReplayInvalidRole

    def is_emitter(self) -> bool:
        """Determine if this replay interface acts as an emitter.
        """
        return self.__role in (ReplayRole.EMITTER, ReplayRole.BOTH)

    def is_receiver(self) -> bool:
        """Determine if this replay interface acts as a receiver.
        """
        return self.__role in (ReplayRole.RECEIVER, ReplayRole.BOTH)

    def prepare(self, **kwargs):
        """Prepare the replay interface.

        This method is called with a set of named parameters depending on the
        parameters declared in the replay configuration.
        """

    def send_packet(self, packet):
        """This method is called to replay a packet from a PCAP file.
        """



"""
WHAD analyzer

This utility processes incoming packets and feed one or more traffic analyzers
with them, extracting and interpreting data based on various supported protocols.
"""
from inspect import Parameter
import sys
from time import sleep
import logging
from typing import List, Union, Tuple

from prompt_toolkit import print_formatted_text, HTML
from scapy.packet import Packet
from scapy.config import conf
from scapy.themes import BrightTheme

from whad.common.analyzer import InvalidParameter
from whad.scapy.layers import *
from whad.cli.app import CommandLineDualOutputApp, run_app
from whad.device.unix import UnixConnector, UnixSocketServer
from whad.device import Bridge
from whad.hub import ProtocolHub
from whad.cli.ui import error, success, info, display_packet, format_analyzer_output
from whad.tools.utils import get_analyzers

logger = logging.getLogger(__name__)
#logging.basicConfig(level=logging.DEBUG)

class WhadAnalyzePipe(Bridge):
    """
    WHAD analyzer main pipe.

    This pipe processes outbound packets.
    """

    def __init__(self, input_connector, output_connector, processing_function):
        super().__init__(input_connector, output_connector)
        self.processing_function = processing_function

    def on_outbound(self, message):
        """Process outbound messages.

        Outbound packets are packets coming from our input connector,that need to be
        forwarded as packets to the next tool.
        """
        if hasattr(message, "to_packet"):
            pkt = message.to_packet()
            if pkt is not None:
                pkts = self.processing_function(pkt, piped=True)
                if pkts is not None:
                    for forwarded in pkts:
                        msg = message.from_packet(forwarded)
                        super().on_outbound(msg)
        else:
            logger.debug("[wfilter][input-pipe] forward default outbound message %s", message)
            # Forward other messages
            super().on_outbound(message)


def display_analyzers(analyzers: dict):
    """Show available analyzers.
    """
    for domain, analyzers_list in analyzers.items():
        print_formatted_text(HTML(
            f"<b><ansicyan>Available analyzers: </ansicyan> {domain}</b>"
        ))
        for analyzer_name, analyzer_class in analyzers_list.items():
            # Retrieve parameters for analyzer class and format them if some are defined.
            analyzer_params = analyzer_class.PARAMETERS
            output=", ".join(analyzer_class().output.keys())
            print_formatted_text(HTML(
                f"  <b>- {analyzer_name}</b> : {output}"
            ))
            parameters = []
            if len(analyzer_params.keys()) > 0:
                for pname, pvalue in analyzer_params.items():
                    print_formatted_text(HTML(
                        f"      <ansiblue>parameter</ansiblue> {pname} <i>(default: \"{pvalue}\")</i>"
                    ))

        print()

class WhadAnalyzeApp(CommandLineDualOutputApp):
    """
    Main `wanalyze` CLI application class.
    """

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD analyze tool',
            interface=False,
            commands=False,
            input=CommandLineDualOutputApp.INPUT_WHAD,
            output=CommandLineDualOutputApp.OUTPUT_STANDARD
        )

        self.add_argument(
            'analyzer',
            help='Analyzer to use',
            nargs="*"
        )


        self.add_argument(
            '--trigger',
            dest='trigger',
            default=False,
            action="store_true",
            help='show when an analyzer is triggered'
        )

        self.add_argument(
            '--json',
            dest='json',
            default=False,
            action="store_true",
            help='serialize output into json format'
        )

        self.add_argument(
            '-p',
            '--packets',
            dest='packets',
            default=False,
            action="store_true",
            help='display packets associated with the analyzer'
        )

        self.add_argument(
            '--label',
            dest='label',
            default=False,
            action="store_true",
            help='display labels before output value'
        )

        self.add_argument(
            '-d',
            '-D',
            '--delimiter',
            dest='delimiter',
            default="\n",
            help='delimiter between outputs'
        )

        self.add_argument(
            '-r',
            '--raw',
            dest='raw',
            action="store_true",
            default=False,
            help='dump output directly to stdout buffer'
        )

        self.add_argument(
            '-l',
            '--list',
            action="store_true",
            help='List of available analyzers'
        )

        self.add_argument(
            '-s',
            '--set',
            dest='ta_params',
            nargs=1,
            action='append',
            metavar='param=value',
            help="set analyzer parameter to value"
        )

        # Initialize analyzers and parameters.
        self.provided_analyzers = []
        self.provided_parameters = {}
        self.selected_analyzers = {}

    def on_packet(self, pkt: Packet, piped: bool = False) -> Union[List[Packet], None]:
        """
        Packet processing callback.

        :param pkt: Packet to process.
        :type pkt: Packet
        :param piped: `True` if packet comes from a pipe, `False` otherwise.
        :type piped: bool, optional
        """
        for analyzer_name, analyzer in self.selected_analyzers.items():
            analyzer.process_packet(pkt)

            if analyzer.triggered and not analyzer._displayed and self.args.trigger and not piped:
                info(analyzer_name + " → " + "triggered")
                analyzer._displayed = True

            if analyzer.completed:
                if self.args.packets and piped:
                    marked_packets = analyzer.marked_packets
                    analyzer.reset()
                    return marked_packets

                if analyzer_name in self.provided_parameters:
                    out = []
                    if not self.args.raw and not self.args.json:
                        for parameter in self.provided_parameters[analyzer_name]:
                            try:
                                out.append(
                                    ((parameter + ": ") if self.args.label else "") +
                                    format_analyzer_output(analyzer.output[parameter],
                                                           mode="human_readable")
                                )
                            except KeyError:
                                error(f"Unknown output parameter {parameter}, ignoring.")
                        print(self.args.delimiter.join(out))
                    elif self.args.raw:
                        for parameter in self.provided_parameters[analyzer_name]:
                            try:
                                sys.stdout.buffer.write(format_analyzer_output(
                                    analyzer.output[parameter], mode="raw"))
                            except KeyError:
                                error(f"Unknown output parameter {parameter}, ignoring.")
                    elif self.args.json:
                        for parameter in self.provided_parameters[analyzer_name]:
                            try:
                                json_value = format_analyzer_output(analyzer.output[parameter],
                                                                    mode="json")
                                if json_value is not None:
                                    if self.args.label:
                                        print("{\"" + parameter + "\": " + json_value + "}")
                                    else:
                                        print(json_value)
                                else:
                                    error((f"Parameter {parameter} not serializable "
                                           f"in JSON, ignoring."))
                            except KeyError:
                                error(f"Unknown output parameter {parameter}, ignoring.")

                    sys.stdout.flush()

                else:
                    if not self.args.json:
                        success("[✓] " + analyzer_name + " → " + "completed")
                        for output_name, output_value in analyzer.output.items():
                            value=format_analyzer_output(output_value, mode="human_readable")
                            if value == "&":
                                value = "&amp;"
                            print_formatted_text(HTML(f"  - <b>{output_name}: </b> {value}"))
                        if self.args.packets:
                            print()
                            print_formatted_text(HTML(
                                f"  - <b>{len(analyzer.marked_packets)} packets analyzed: </b>"
                            ))

                            for packet in analyzer.marked_packets:
                                display_packet(
                                    packet,
                                    show_metadata = self.args.metadata,
                                    format = self.args.format
                                )
                        print()
                    else:
                        out = "{"
                        for output_name, output_value in analyzer.output.items():
                            json_value = format_analyzer_output(output_value,mode="json")
                            out += "\"" + output_name + "\": " + json_value + ", "
                        out = out[:-2] + "}"
                        print(out)
                    analyzer._displayed = False
                analyzer.reset()

        # No packets returned
        return None

    def get_provided_analyzers(self) -> Tuple[list, dict]:
        """Retrieve a list of analyzers with their corresponding parameters.

        :return: list of available analyzers and their associated parameters.
        :rtype: tuple
        """
        available_analyzers = list(get_analyzers(self.args.domain).keys())
        analyzers = []
        parameters = {}
        for analyzer in self.args.analyzer:
            if "." in analyzer:
                name, param = analyzer.split(".")
                if name in available_analyzers:
                    analyzers.append(name)
                    if name in parameters:
                        parameters[name].append(param)
                    else:
                        parameters[name] = [param]
                else:
                    error(f"Unknown analyzer ({name}).")
                    sys.exit(1)
            else:
                if analyzer in available_analyzers:
                    analyzers.append(analyzer)
                else:
                    error(f"Unknown analyzer ({analyzer}).")
                    sys.exit(1)
        return list(set(analyzers)), parameters

    def run(self):
        # Launch pre-run tasks
        self.pre_run()

        # Parse analyzer parameters passed by the user (--set param=value)
        config_params = {}
        if self.args.ta_params is not None:
            for param in self.args.ta_params:
                if '=' in param[0]:
                    pname, pvalue = param[0].split('=')[:2]
                    if pname != '':
                        config_params[pname] = pvalue

        if self.args.list:
            analyzers = get_analyzers()
            display_analyzers(analyzers)

        try:
            if self.is_piped_interface():
                interface = self.input_interface
            else:
                interface = self.interface

            if interface is not None:
                if not self.args.nocolor:
                    conf.color_theme = BrightTheme()

                # Configure the current domain
                ProtocolHub.set_domain(self.args.domain)

                #parameters = self.args.__dict__
                connector = UnixConnector(interface)
                #for parameter_name, parameter_value in parameters.items():
                #    connector.add_parameter(parameter_name, parameter_value)

                self.provided_analyzers, self.provided_parameters = self.get_provided_analyzers()
                self.selected_analyzers = {}
                for name, clazz in get_analyzers(self.args.domain).items():
                    if name in self.provided_analyzers or len(self.provided_analyzers) == 0:
                        # Pick parameters required by the analyzer from our current configured params
                        params = {}
                        for pname, pvalue in config_params.items():
                            if pname in clazz.PARAMETERS:
                                params[pname] = pvalue

                        # Instantiate the requested traffic analyzer with its optional params
                        self.selected_analyzers[name] = clazz(**params)
                        self.selected_analyzers[name]._displayed = False

                connector.domain = self.args.domain
                hub = ProtocolHub(2)
                connector.format = hub.get(self.args.domain).format


                if self.is_stdout_piped() and self.args.packets:
                    unix_server = UnixConnector(UnixSocketServer(parameters={
                        'domain': self.args.domain,
                        'format': self.args.format,
                        'metadata' : self.args.metadata
                    }))


                    while not unix_server.device.opened:
                        if unix_server.device.timedout:
                            return
                        sleep(0.1)

                    # Create our packet bridge
                    logger.info("[wanalyze] Starting our output pipe")
                    WhadAnalyzePipe(connector, unix_server, self.on_packet).wait()
                else:
                    connector.on_packet = self.on_packet
                    # Unlock connector
                    connector.unlock()

                    # Wait for the associated interface to disconnect
                    connector.join()
        except InvalidParameter as param_err:
            # Invalid parameter provided
            self.error(f"An invalid value has been provided for parameter '{param_err.parameter}' ({param_err.value}) !")
        except KeyboardInterrupt:
            # Launch post-run tasks
            self.post_run()


def wanalyze_main():
    """Launcher for wanalyze CLI application.
    """
    app = WhadAnalyzeApp()
    run_app(app)




"""WHAD Protocol BLE mode messages abstraction layer.
"""
from typing import Optional
from whad.protocol.whad_pb2 import Message
from whad.protocol.ble.ble_pb2 import CentralModeCmd, StartCmd as BleStartCmd, StopCmd as BleStopCmd
from whad.hub.message import pb_bind, PbFieldBytes, PbMessageWrapper, PbFieldBool, PbFieldInt
from whad.hub.ble import BleDomain

@pb_bind(BleDomain, 'scan_mode', 1)
class ScanMode(PbMessageWrapper):
    """BLE scan mode message class
    """
    active = PbFieldBool('ble.scan_mode.active_scan')

@pb_bind(BleDomain, 'adv_mode', 1)
class AdvMode(PbMessageWrapper):
    """BLE advertising mode message class
    """
    scan_data = PbFieldBytes('ble.adv_mode.scan_data')
    scanrsp_data = PbFieldBytes('ble.adv_mode.scanrsp_data')

@pb_bind(BleDomain, 'central_mode', 1)
class CentralMode(PbMessageWrapper):
    """BLE advertising mode message class
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.ble.central_mode.CopyFrom(CentralModeCmd())

@pb_bind(BleDomain, 'periph_mode', 1)
class PeriphMode(PbMessageWrapper):
    """BLE advertising mode message class
    """
    scan_data = PbFieldBytes('ble.periph_mode.scan_data')
    scanrsp_data = PbFieldBytes('ble.periph_mode.scanrsp_data')

    def get_adv_data(self) -> Optional[bytes]:
        """Retrieve advertising data."""
        return self.get_field_value(PeriphMode.scan_data)

    def get_scan_data(self) -> Optional[bytes]:
        """Retrieve scan response data, if set."""
        return self.get_field_value(PeriphMode.scanrsp_data)

@pb_bind(BleDomain, 'start', 1)
class BleStart(PbMessageWrapper):
    """BLE start mode message class
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.ble.start.CopyFrom(BleStartCmd())

@pb_bind(BleDomain, 'stop', 1)
class BleStop(PbMessageWrapper):
    """BLE stop mode message class
    """

    def __init__(self, message: Message = None):
        super().__init__(message=message)
        self.message.ble.stop.CopyFrom(BleStopCmd())

@pb_bind(BleDomain, 'encryption', 1)
class SetEncryption(PbMessageWrapper):
    """BLE SetEncryption message class
    """
    conn_handle = PbFieldInt('ble.encryption.conn_handle')
    enabled = PbFieldBool('ble.encryption.enabled')
    ll_key = PbFieldBytes('ble.encryption.ll_key')
    ll_iv = PbFieldBytes('ble.encryption.ll_iv')
    key = PbFieldBytes('ble.encryption.key')
    rand = PbFieldBytes('ble.encryption.rand')
    ediv = PbFieldBytes('ble.encryption.ediv')



"""This module provides specific classes to set-up a Unifying proxy.

The class :class:`LinkLayerProxy` provides a link-layer level Unifying proxy that
implements different mechanisms to setup a Man-in-the-Middle in an Unifying network.

Once established, the class will act as a proxy, allowing to block, modify or formward
packets transmitted to the dongle.
"""
from whad.unifying.connector import Unifying, Mouse, Dongle
from whad.esb.stack import ESBStack
from whad.exceptions import WhadDeviceNotFound
from whad.unifying.stack import UnifyingApplicativeLayer, UnifyingRole
from whad.scapy.layers.esb import ESB_Hdr, ESB_Payload_Hdr
from whad.scapy.layers.unifying import Logitech_Unifying_Hdr, Logitech_Keepalive_Payload, \
    Logitech_Set_Keepalive_Payload
from random import choice
from enum import IntEnum
from time import sleep
# Logging
import logging
logger = logging.getLogger(__name__)

def pick_random_unifying_channel():
    """
    Returns a randomly chosen Unifying channel.
    """
    channels = [5, 8, 11, 14, 17, 20, 29, 32, 35, 38, 41, 44, 47, 56, 59, 62, 65, 68, 71, 74]
    return choice(channels)


class LowLevelDongle(Dongle):
    """Link-layer only PDU implementation

    This class is used by the :class:`whad.unifying.tools.proxy.LinkLayerProxy` class
    to provide a Link-Layer dongle.
    """

    def __init__(self, device, proxy=None):
        super().__init__(device)
        self.__proxy = proxy
        self.__other_half = None

    def set_other_half(self, other_half):
        self.__other_half = other_half


    def on_pdu(self, packet):
        print("[dongle] received PDU: " + repr(packet) +("(channel = %d)" % packet.metadata.channel))

        # Apply potential modifications from proxy and forward to other half
        if self.__other_half is not None:
            if self.__proxy is not None:
                pdu = self.__proxy.on_pdu(packet)

                if pdu is not None:
                    self.__other_half.forward_pdu(pdu)
                else:
                    self.__other_half.forward_pdu(pdu)
        super().on_pdu(packet)

    def forward_pdu(self, pdu):
        """Forward a PDU to the node, if any

        :param Packet pdu: PDU to send to the node
        """
        try:
            payload = pdu[Logitech_Unifying_Hdr:][1:]
            return self.stack.app.prepare_message(payload)
        except IndexError:
            return False

class LowLevelNode(Mouse):
    """Link-layer only Node implementation

    This class is used by the :class:`whad.unifying.tools.proxy.LinkLayerProxy` class
    to provide a Link-Layer node.

    """
    def __init__(self, device, proxy=None):
        super().__init__(device)
        self.__proxy = proxy
        self.__other_half = None

    def set_other_half(self, other_half):
        self.__other_half = other_half

    def on_pdu(self, packet):

        # We don't forward packet if it is a acknowledgment with an empty payload
        # because the other half will automatically acknowledge packets
        super().on_pdu(packet)

        if len(bytes(packet[ESB_Payload_Hdr:])) == 0:
            self.__proxy.on_ack()
            return

        print("[node] received packet " + repr(packet)  +("(channel = %d)" % packet.metadata.channel))

        # Apply potential modifications from proxy and forward to other half
        if self.__other_half is not None:
            if self.__proxy is not None:
                pdu = self.__proxy.on_pdu(packet)
                if pdu is not None:
                    self.__other_half.forward_pdu(pdu)
                else:
                    self.__other_half.forward_pdu(pdu)

    def forward_pdu(self, pdu):
        """Forward a PDU to the dongle, if any

        :param Packet pdu: PDU to send to the dongle
        """
        try:
            payload = pdu[Logitech_Unifying_Hdr:][1:]
            payload.show()
            return self.stack.app.prepare_message(payload)
            self.send(pdu)

        except IndexError:
            return False


class LinkLayerProxy(object):
    """This class implements a Unifying proxy that relies on two Unifying-compatible
    WHAD devices to create a real Unifying device that will proxify all the link-layer
    traffic to another device.

    By default, it will imitate the behaviour of a dongle on a first Unifying channel and
    the behaviour of a Mouse / Keyboard on a second channel.
    """

    def __init__(self, proxy=None, target=None, address=None, desync=False, proxy_channel=None, target_channel=None):
        """
        :param Unifying proxy: Unifying device to use as dongle
        :param Unifying target: Unifying device to use as node
        :param str address: Unifying address of target device
        :param bool desync: desynchronize an existing connection
        :param int proxy_channel: proxy channel to use
        :param int target_channel: target channel to use
        """
        if proxy is None or target is None or address is None:
            raise WhadDeviceNotFound

        # Save both devices
        self.__proxy = proxy
        self.__dongle = None
        self.__target = target
        self.__node = None
        self.__target_addr = address
        self.__proxy_channel = None
        self.__target_channel = None

        self.__desync = desync

        self.__ack_counter = 0
        # Callbacks
        self.__callbacks = []

    @property
    def target(self):
        return self.__node


    @property
    def proxy(self):
        return self.__dongle



    def close(self):
        if self.__proxy is not None:
            self.__proxy.close()
        if self.__target is not None:
            self.__target.close()


    def on_ack(self):
        self.__ack_counter += 1

    def on_pdu(self, pdu):

        if hasattr(pdu, "button_mask"):
            if pdu.button_mask == 0x01:
                pdu.button_mask = 2
            elif pdu.button_mask == 0x02:
                pdu.button_mask = 1
        return pdu

    def start(self):
        """Start proxy

        The proxy device will be set as a dongle
        """

        # First, connect our central device to our target device
        print('create low-level dongle device ...')
        self.__dongle = LowLevelDongle(self.__proxy, self)

        # If proxy channel is provided, use it, otherwise pick one Unifying channel randomly
        if self.__proxy_channel is not None:
            self.__dongle.channel = self.__proxy_channel
        else:
            self.__dongle.channel = pick_random_unifying_channel()

        self.__dongle.address = self.__target_addr

        print('create low-level node device ...')
        self.__node = LowLevelNode(self.__target, self)

        # If target channel is provided, use it, otherwise pick one Unifying channel randomly
        if self.__target_channel is not None:
            self.__node.channel = self.__target_channel
        else:
            self.__node.channel = pick_random_unifying_channel()

        self.__node.address = self.__target_addr


        self.__node.set_other_half(self.__dongle)
        self.__dongle.set_other_half(self.__node)

        if self.__desync:
            self.desync()
        else:
            # Start the node, lock the channel
            self.__node.start()
            self.__node.lock()

            # Start the dongle
            self.__dongle.start()



    def desync(self):
        self.__node.start()
        self.__dongle.start()

        while True:
            print("Synchronizing ...")
            #self.__node.start()
            while not self.__node.synchronize():
                sleep(0.1)
            self.__dongle.channel = self.__node.channel
            while self.__node.channel == self.__dongle.channel:
                self.__node.channel = pick_random_unifying_channel()
            #self.__node.stop()

            self.__node.lock()
            #self.__node.start()
            #self.__dongle.start()
            self.__ack_counter = 0
            self.__dongle.auto(False)
            print("Injecting keepalive...")
            self.__dongle.send(ESB_Hdr()/ESB_Payload_Hdr()/Logitech_Unifying_Hdr()/Logitech_Set_Keepalive_Payload(timeout=1))
            self.__dongle.send(ESB_Hdr()/ESB_Payload_Hdr()/Logitech_Unifying_Hdr()/Logitech_Keepalive_Payload(timeout=1))
            self.__dongle.auto(True)
            sleep(5)
            if self.__ack_counter > 5:
                return True

            self.__node.unlock()
            #self.__node.stop()
            #self.__dongle.stop()

    def stop(self):
        """
        Stop the proxy.
        """
        if self.__dongle is not None:
            self.__dongle.stop()
        if self.__node is not None:
            self.__node.stop()



"""
BLE Scanning device database
============================

This module provides a database that keeps track of discovered devices,
:class:`whad.ble.scanning.AdvertisingDevicesDB`. Discovered devices information
are handled in :class:`whad.ble.scanning.AdvertisingDevice`.
"""
from time import time
from typing import Optional

from scapy.layers.bluetooth4LE import BTLE_ADV_IND, BTLE_ADV_NONCONN_IND, \
    BTLE_SCAN_RSP, BTLE_ADV

from whad.hub.ble import BDAddress
from whad.ble.profile.advdata import AdvDataFieldList, AdvCompleteLocalName, \
    AdvDataError, AdvDataFieldListOverflow, AdvShortenedLocalName

class AdvertisingDevice:
    """Store information about a device:

    * Received Signal Strength Indicator (RSSI)
    * Address type (public or random)
    * Advertising data
    * Scan response data
    * Type of advertising PDU
    * Connectable information
    """

    SCAN_RSP_TIMEOUT = 0.5

    def __init__(self, rssi, address_type, bd_address, adv_data, rsp_data=None,
                 undirected=True, connectable=True):
        """Instantiate an AdvertisingDevice.

        :param  rssi:           Received Signal Strength Indicator
        :type   rssi:           float
        :param  address_type:   Address type
        :type   address_type:   int
        :param  bd_address:     Bluetooth device address
        :type   bd_address:     str
        :param  adv_data:       Advertising data
        :type   adv_data:       bytes
        :param  rsp_data:       Scan response data
        :type   rsp_data:       bytes, optional
        :param  undirected:     ``True`` if advertising PDUs are undirected, `False` otherwise
        :type   undirected:     bool, optional
        :param  connectable:    ``True`` if device accepts connection, `False` otherwise
        :type   connectable:    bool, optional
        """
        self.__address_type = address_type
        self.__bd_address = bd_address
        self.__adv_data = adv_data
        self.__rsp_data = rsp_data
        self.__rssi = rssi
        self.__got_scan_rsp = False
        self.__undirected = undirected
        self.__connectable = connectable
        self.__scanned = False
        self.__reported = False
        self.__timestamp = time()
        self.__last_seen = self.__timestamp

    @property
    def address(self) -> str:
        """Device BD address.
        """
        return str(self.__bd_address)

    @property
    def address_type(self) -> int:
        """Device address type.
        """
        return self.__address_type

    @property
    def rssi(self) -> float:
        """Device RSSI.
        """
        return self.__rssi

    @property
    def adv_records(self) -> bytes:
        """Advertising records.
        """
        return self.__adv_data

    @property
    def scan_rsp_records(self) -> bytes:
        """Scan response records.
        """
        return self.__rsp_data

    @property
    def ad_records(self) -> AdvDataFieldList:
        """Combined advertising and scan response records.
        """
        out = AdvDataFieldList()
        for record in self.__adv_data:
            out.add(record)
        if self.__rsp_data is not None:
            for record in self.__rsp_data:
                out.add(record)
        return out

    @property
    def got_scan_rsp(self) -> bool:
        """Received a scan response from device.
        """
        return self.__got_scan_rsp

    @property
    def name(self) -> str:
        """Device complete or short name.
        """
        # Do we have a name ?
        complete_name = None
        short_name = None
        for record in self.ad_records:
            if isinstance(record, AdvShortenedLocalName):
                short_name = record.name.decode('utf-8')
            elif isinstance(record, AdvCompleteLocalName):
                complete_name = record.name.decode('utf-8')

        # Return discovered name (if any)
        if complete_name is not None:
            return complete_name

        if short_name is not None:
            return short_name

        return None

    @property
    def scanned(self) -> bool:
        """Device scanned status
        """
        return self.__scanned

    @property
    def timestamp(self) -> float:
        """Device discovery timestamp.
        """
        return self.__timestamp

    @property
    def last_seen(self) -> float:
        """Device last seen timestamp.
        """
        return self.__last_seen

    @property
    def reported(self) -> bool:
        """Reported status
        """
        return self.__reported

    @property
    def connectable(self) -> bool:
        """Connectable status.
        """
        return self.__connectable

    def __eq__(self, other):
        """Compare this device to another. Devices are considered unique
        based on their BD address.
        """
        return self.address == other.address

    def __repr__(self):
        """Show device information.
        """
        # Do we have a name ?
        complete_name = None
        short_name = None
        for record in self.ad_records:
            if isinstance(record, AdvShortenedLocalName):
                try:
                    short_name = record.name.decode('utf-8')
                except UnicodeDecodeError:
                    short_name = record.name.decode('latin1')
            elif isinstance(record, AdvCompleteLocalName):
                try:
                    complete_name = record.name.decode('utf-8')
                except UnicodeDecodeError:
                    complete_name = record.name.decode('latin1')

        # Pick the best name
        if complete_name:
            name = f'name:"{complete_name}"'
        elif short_name:
            name = f'name:"{short_name}"'
        else:
            name = ''

        # Display address type
        if self.__address_type == 0:
            addrtype = '[PUB]'
        else:
            addrtype = '[RND]'

        # Generate device summary
        return f'[{self.__rssi:04d} dBm] {addrtype} {self.__bd_address} {name}'


    def seen(self):
        """Mark device as seen (update last_seen value with current time).
        """
        self.__last_seen = time()

    def mark_reported(self):
        """Mark device as reported (avoid duplicates)
        """
        self.__reported = True

    def update(self, rssi : float = None, adv_data : bytes = None):
        """Update device RSSI and advertising data and check for scan response
        timeout.

        :param  rssi: New RSSI value.
        :type   rssi: float
        :param  adv_data: New advertising data
        :type   adv_data: bytes
        """
        if rssi is not None:
            self.__rssi = rssi

        if adv_data is not None:
            self.__adv_data = adv_data

        # Update scanned status if required
        if not self.__scanned:
            if (time() - self.__timestamp) > self.SCAN_RSP_TIMEOUT:
                self.__scanned = True


    def set_scan_rsp(self, scan_rsp):
        """Update device advertisement data.

        :param  scan_rsp:   Raw scan response.
        :type   scan_rsp:   bytes
        """
        if not self.__got_scan_rsp:
            self.__rsp_data = scan_rsp
            self.__got_scan_rsp = True
            self.__scanned = True


class AdvertisingDevicesDB:
    """Bluetooth Low Energy devices database.

    This class stores information about discovered devices.
    """

    def __init__(self):
        self.reset()


    def reset(self):
        """Remove database content.
        """
        self.__db = {}


    def find_device(self, address: str) -> AdvertisingDevice:
        """Find a device based on its BD address.

        :param      address: Device BD address
        :type       address: str
        :return:    Device if found, `None` otherwise.
        :rtype:     :class:`whad.ble.scanning.AdvertisingDevice`
        """
        device = None
        address = address.lower()
        if address in self.__db:
            device = self.__db[address]
        return device

    def register_device(self, device, update: bool = False) -> bool:
        """Register or update a device.

        :param  device: Device to register.
        :type   device: :class:`whad.ble.scanning.AdvertisingDevice`
        """
        if device.address not in self.__db:
            self.__db[device.address] = device
            return True
        else:
            dev = self.__db[device.address]
            dev.seen()
            if dev.rssi != device.rssi:
                dev.update(rssi = device.rssi)
                return update
        return False


    def __apply_scan_rsp_timeout(self):
        """Check every device to determine if our scan request has timed out.
        """
        for _, device in self.__db.items():
            if not device.scanned:
                device.update()
            if device.scanned and not device.reported:
                device.mark_reported()
                yield device


    def on_device_found(self, rssi, adv_packet, filter_addr: Optional[str] = None, updates: bool = False):
        """Device advertising packet or scan response received.

        Parse the incoming packet and handle device appropriately.

        :param  rssi:           Received Signal Strength Indicator
        :type   rssi:           float
        :param  adv_packet:     Advertising packet
        :type   adv_packet:     :class:`scapy.packet.Packet`
        :param  filter_addr:    BD address to filter
        :type   filter_addr:    str
        """
        devices = []
        addr_type = adv_packet.getlayer(BTLE_ADV).TxAdd

        if adv_packet.haslayer(BTLE_ADV_IND):
            bd_address = BDAddress(adv_packet[BTLE_ADV_IND].AdvA)
            try:
                adv_data = b''.join([ bytes(record) for record in adv_packet[BTLE_ADV_IND].data])
                adv_list = AdvDataFieldList.from_bytes(adv_data)
                device = AdvertisingDevice(
                    rssi,
                    addr_type,
                    bd_address,
                    adv_list
                )


                # Register device if it matches our criterias
                if filter_addr is not None and filter_addr.lower() == str(bd_address).lower():
                    if self.register_device(device, update=updates) and updates:
                        devices.append(device)
                elif filter_addr is None:
                    if self.register_device(device, update=updates) and updates:
                        devices.append(device)
            except AdvDataError:
                pass
            except AdvDataFieldListOverflow:
                pass

        elif adv_packet.haslayer(BTLE_ADV_NONCONN_IND):
            try:
                bd_address = BDAddress(adv_packet[BTLE_ADV_NONCONN_IND].AdvA)
                adv_data = b''.join([ bytes(record)
                                     for record in adv_packet[BTLE_ADV_NONCONN_IND].data])
                adv_list = AdvDataFieldList.from_bytes(adv_data)
                device = AdvertisingDevice(
                    rssi,
                    addr_type,
                    bd_address,
                    adv_list,
                    connectable=False
                )

                # Register device if it matches our criterias
                if filter_addr is not None and filter_addr.lower() == str(bd_address).lower():
                    if self.register_device(device, update=updates) and updates:
                        devices.append(device)
                elif filter_addr is None:
                    if self.register_device(device, update=updates) and updates:
                        devices.append(device)

            except AdvDataError:
                pass
            except AdvDataFieldListOverflow:
                pass

        elif adv_packet.haslayer(BTLE_SCAN_RSP):
            try:
                bd_address = BDAddress(adv_packet[BTLE_SCAN_RSP].AdvA)
                adv_data = b''.join([ bytes(record) for record in adv_packet[BTLE_SCAN_RSP].data])
                adv_list = AdvDataFieldList.from_bytes(adv_data)
                if str(bd_address) in self.__db:
                    device = self.__db[str(bd_address)]
                    if not device.got_scan_rsp:
                        device.set_scan_rsp(adv_list)
                        if (filter_addr is not None and filter_addr.lower() == str(bd_address).lower()) or updates:
                            devices.append(device)
            except AdvDataError:
                pass
            except AdvDataFieldListOverflow:
                pass

        # Check if some devices scan response timeout is reached
        for device in self.__apply_scan_rsp_timeout():
            if device not in devices:
                devices.append(device)

        # If asked for updates, send all updated devices except
        # those already processed
        if not updates:
            filter(lambda x: x.reported == False, devices)

        return devices



"""WHAD Protocol PHY schedule packet messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from ..message import pb_bind, PbFieldInt, PbFieldBytes, PbFieldBool, PbMessageWrapper
from . import PhyDomain

from .timestamp import Timestamp

@pb_bind(PhyDomain, 'sched_send', 1)
class SchedulePacket(PbMessageWrapper):
    """PHY schedule packet send message
    """

    packet = PbFieldBytes('phy.sched_send.packet')
    timestamp = PbFieldInt('phy.sched_send.timestamp')

@pb_bind(PhyDomain, 'sched_pkt_rsp', 1)
class SchedulePacketResponse(PbMessageWrapper):
    """PHY schedule packet response message
    """

    id = PbFieldInt('phy.sched_pkt_rsp.id')
    full = PbFieldBool('phy.sched_pkt_rsp.full', optional=True)

@pb_bind(PhyDomain, 'sched_pkt_sent', 1)
class ScheduledPacketSent(PbMessageWrapper):
    """PHY schedule packet sent notification message
    """

    id = PbFieldInt('phy.sched_pkt_sent.id')



"""
BLE GATT Characteristic Model
=============================
"""
from re import I
from struct import pack, unpack
from typing import Union, Type, Optional, List, Iterator

from whad.ble.stack.att.constants import BleAttProperties, SecurityAccess
from whad.ble.profile.attribute import Attribute, UUID, get_uuid_alias
from whad.ble.exceptions import InvalidHandleValueException
from whad.ble.utils.clues import CluesDb

class desc_type:

    def __init__(self, uuid: UUID):
        self.__uuid = uuid

    def __call__(self, cls):
        Descriptor.register_type(self.__uuid, cls)
        return cls

class Properties:
    """Generic characteristic properties.
    """
    BROADCAST = 0x01
    READ = 0x02
    WRITE_WITHOUT_RESPONSE = 0x04
    WRITE = 0x08
    NOTIFY = 0x10
    INDICATE = 0x20
    AUTH_SIGNED_WRITES = 0x40
    EXTENDED_PROPERTIES = 0x80

class CharacteristicProperties(Properties):
    """Old name, defined for compatibility"""

class Descriptor(Attribute):
    """BLE Characteristic descriptor
    """
    desc_types = {}

    def __init__(self, uuid: UUID, handle: int = 0, value: bytes = b'', characteristic: Optional['Characteristic'] = None):
        super().__init__(uuid=uuid,handle=handle,value=value)
        self.__characteristic = characteristic

    @property
    def characteristic(self) -> Optional['Characteristic']:
        """Parent characteristic
        """
        return self.__characteristic

    @characteristic.setter
    def characteristic(self, charac: 'Characteristic'):
        """Set characteristic."""
        self.__characteristic = charac

    @property
    def uuid(self):
        """Descriptor UUID
        """
        return self.type_uuid

    @property
    def name(self):
        """Descriptor name
        """
        # Search Bluetooth known UUIDs
        alias = get_uuid_alias(self.type_uuid)
        if alias is not None:
            return f"{alias} (0x{self.type_uuid})"

        # No alias
        return str(self.type_uuid)

    @staticmethod
    def register_type(uuid: UUID, cls):
        """Register descriptor type (associate a descriptor UUID with the
        corresponding Python class (must inherit from Descriptor)
        """
        uuid_value = uuid.value()
        if uuid_value not in Descriptor.desc_types:
            if issubclass(cls, Descriptor):
                Descriptor.desc_types[uuid_value] = cls

    @staticmethod
    def get_type_uuid(desc_cls) -> Optional[UUID]:
        """Find the type UUID corresponding to a given registered descriptor's
        class.

        :return: Descriptor's type UUID if found, `None` otherwise.
        :rtype: UUID
        """
        for desc_type,desc_cls in Descriptor.desc_types.items():
            if desc_cls == desc_cls:
                return UUID(desc_type)

        # Not found
        return None

    @staticmethod
    def from_uuid(characteristic, handle: int, uuid: UUID, value: bytes):
        """Create an instance of a descriptor based on the provided UUID and
        descriptor value.

        :param uuid: Descriptor UUID
        :type  uuid: UUID
        :param value: Descriptor value
        :type  value: bytes
        :return: Instance of the corresponding descriptor
        :rtype: Descriptor
        """
        uuid_value = uuid.value()
        if uuid_value in Descriptor.desc_types:
            cls = Descriptor.desc_types[uuid_value]
            return cls.from_value(characteristic, handle, value)

        # Cannot find any class matching the provided UUID, return a generic
        # descriptor.
        return Descriptor(uuid, handle, value, characteristic)

    @classmethod
    def _build(cls, instance: Type['Descriptor']):
        """Build a new descriptor based on current template."""
        return cls(
            instance.uuid,
            0,
            instance.value,
            characteristic=instance.characteristic
        )

    def build(self):
        return self.__class__._build(self)

class CharacteristicDescriptor(Descriptor):
    """Old class defining a characteristic descriptor, kept for backward compatibility.

    .. deprecated:: 1.3.0
        Renamed for clarity purpose ('Characteristic' in this class name has been removed
        as this class is already defined in the *characteristic* module), please use
        the new :class:`~whad.ble.profile.characteristic.Descriptor` class instead.
    """

@desc_type(UUID(0x2902))
class ClientCharacteristicConfig(Descriptor):
    """Client Characteristic Configuration Descriptor
    """

    def __init__(self, handle: int = 0, notify: bool = False, indicate: bool = False,
                 characteristic: Optional['Characteristic'] = None):
        """Instantiate a Client Characteristic Configuration Descriptor

        :param bool notify: Set to True to get the corresponding characteristic notified on change
        :param bool indicate: Set to True to get the corresponding characteristic
                              indicated on change
        """
        value = 0
        if notify:
            value |= 0x0001
        if indicate:
            value |= 0x0002
        super().__init__(UUID(0x2902), handle=handle, value=pack('<H', value), characteristic=characteristic)

    @property
    def config(self):
        """CCCD configuration
        """
        return unpack('<H', super().value)[0]

    @config.setter
    def config(self, val):
        """Update CCCD configuration
        """
        super().value = pack('<H', val)

    @staticmethod
    def from_value(characteristic, handle, value):
        """Create a ClientCharacteristicConfig instance from the provided
        handle and value, and tie it to a specific characteristic.
        """
        # Create our CCCD object
        if len(value) >= 1:
            v = value[0]
        else:
            v = 0x00
        cccd = ClientCharacteristicConfig(handle,(v & 0x0001) != 0, (v & 0x0002) != 0, characteristic)
        # Set its value
        cccd.value = value

        return cccd

    @classmethod
    def _build(cls, instance: 'ClientCharacteristicConfig'):
        """Build a new descriptor based on current template."""
        return cls(
            0,
            (instance.config & 0x0001) != 0,
            (instance.config & 0x0002) != 0,
            characteristic=instance.characteristic
        )

class ReportReference(Descriptor):
    """Report Reference descriptor, used in HID profile
    """

    def __init__(self, handle: int = 0, characteristic: Optional['Characteristic'] = None):
        """Instantiate a Report Reference Descriptor

        :param bool notify: Set to True to get the corresponding characteristic notified on change
        :param bool indicate: Set to True to get the corresponding characteristic
                              indicated on change
        """
        super().__init__(
            uuid=UUID(0x2908),
            handle=handle,
            value=b'\x01\x01',
            characteristic=characteristic
        )

    @classmethod
    def _build(cls, instance: 'ReportReference'):
        """Build a new descriptor based on current template."""
        return cls(
            0,
            characteristic=instance.characteristic
        )

class ReportReferenceDescriptor(ReportReference):
    """Old report reference descriptor class, kept for backward compatibility.


    .. deprecated:: 1.3.0
        Use the new :class:`~whad.ble.profile.characteristic.ReportReference` class that defines
        an HID ReportReference descriptor.
    """

@desc_type(UUID(0x2901))
class UserDescription(Descriptor):
    """Characteristic user description descriptor.

    This descriptor specifies a textual description of the characteristic it is attached to. This description
    is exposed by the GATT server and can be read by a GATT client.
    """
    def __init__(self, handle: int = 0, description: str = '', characteristic: Optional['Characteristic'] = None):
        """Instantiate a Characteristic User Description descriptor

        :param description: Set characteristic text description
        :type  description: str
        :param characteristic: Characteristic this descriptor is attached to, not required when defining a GATT model.
        :type  characteristic: Characteristic, optional
        """
        self.__description = description
        super().__init__(
            uuid=UUID(0x2901),
            handle=handle,
            value=description.encode('utf-8'),
            characteristic=characteristic
        )

    @property
    def text(self) -> str:
        """User description
        """
        return self.__description

    @property
    def name(self):
        """Descriptor name

        Clean description as it may contain null chars.
        """
        # End string at the first null char encountered
        if '\x00' in self.__description:
            desc = self.__description[:self.__description.index('\x00')]
        else:
            desc = self.__description

        return f"{super().name}, '{desc}'"

    @staticmethod
    def from_value(characteristic, handle, value):
        """Create CUD descriptor from value
        """
        return UserDescription(
             handle, description=value.decode('utf-8'), characteristic=characteristic
        )

    @classmethod
    def _build(cls, instance: 'UserDescription'):
        """Build a new descriptor based on current template."""
        return cls(
            0,
            instance.value.decode('utf-8'),
            characteristic=instance.characteristic
        )

class CharacteristicUserDescriptionDescriptor(UserDescription):
    """Old name of the UserDescription descriptor."""

class CharacteristicValue(Attribute):
    """Characteristic value attribute.
    """

    def __init__(self, uuid, handle: int = 0, value: bytes = b'', characteristic: Optional['Characteristic'] = None):
        super().__init__(uuid=uuid, handle=handle, value=value)
        self.__characteristic = characteristic

    @property
    def characteristic(self):
        """Associated characteristic
        """
        return self.__characteristic

    @property
    def uuid(self):
        """Attribute UUID
        """
        return self.type_uuid

class Characteristic(Attribute):
    """BLE Characteristic
    """

    # Properties
    BROADCAST = 0x01
    READ = 0x02
    WRITE_WITHOUT_RESPONSE = 0x04
    WRITE = 0x08
    NOTIFY = 0x10
    INDICATE = 0x20
    AUTH_SIGNED_WRITES = 0x40
    EXTENDED_PROPERTIES = 0x80

    def __init__(self, uuid: UUID, handle: int = 0, end_handle: int = 0, value: bytes = b'',
                 properties: int = 0, permissions: Optional[List[str]] = None, notify: bool = False,
                 indicate: bool = False, required: bool = False, description: Optional[str] = None, security:
                 Optional[SecurityAccess] = None, descriptors: List[Descriptor] = []):
        """Instantiate a BLE characteristic object

        :param uuid: 16-bit or 128-bit UUID
        :param int handle: Handle value
        :param bytes value: Characteristic value
        :param int perms: Permissions
        """
        # Create our parent attribute (type UUID of 0x2803, and value containing our properties, value handle and UUID)
        super().__init__(uuid=UUID(0x2803), handle=handle,
                         value=pack('<BH', properties & 0xff, handle+1)+uuid.to_bytes())

        # Set end handle: if 0, then it is considered as not set by the caller and set by
        # default to the characteristic's handle value.
        if end_handle == 0:
            self.__end_handle = handle
        else:
            self.__end_handle = end_handle

        # Set characteristic UUID
        self.__charac_uuid = uuid

        # Set characteristic alias
        self.__alias = None

        # Notification and indication callbacks
        self.__notification_callback = None
        self.__indication_callback = None

        # Add a characteristic value attribute
        self.__value_handle = handle + 1
        self.__value = CharacteristicValue(uuid, self.__value_handle, value, self)
        self.__end_handle = self.__value_handle

        # Set characteristic properties
        self.__properties = properties

        # Update properties according to defined permissions, if any (only used when object is considered a template)
        if permissions is not None:
            perms = list(map(lambda x: x.lower().strip(), permissions))
            if 'read' in perms:
                self.__properties |= Characteristic.READ
            if 'write' in perms:
                self.__properties |= Characteristic.WRITE
            if 'write_without_response' in perms:
                self.__properties |= Characteristic.WRITE_WITHOUT_RESPONSE
            if 'notify' in perms:
                self.__properties |= Characteristic.NOTIFY
            if 'indicate' in perms:
                self.__properties |= Characteristic.INDICATE

        # If characteristic is set as supporting notifications, define the correct property.
        if notify:
            self.__properties |= Properties.NOTIFY

        # Same for indications.
        if indicate:
            self.__properties |= Properties.INDICATE

        # By default, this characteristic object is instantiated as a standalone characteristic,
        # this means all the handles depending on this characteristic (value and descriptors) will
        # be computed from the characteristic's definition handle.
        self.__service = None

        # When `required` is set, the characteristic is set as *mandatory*. This flag is only used by a
        # service to determine if this characteristic must be present in its parent service.
        self.required = required

        # Security properties
        self.__security = security if security is not None else []

        # List of descriptors.
        self.__descriptors = []

        # Add descriptors if handle is 0 (object used as a template)
        if handle == 0:
            # If this characteristic's handle is not set, we automatically add a ClientCharacteristicConfig
            # descriptor if its properties say it supports indication and/or notification.
            if (self.can_indicate() or self.can_notify()):
                self.add_descriptor(ClientCharacteristicConfig(
                    characteristic=self
                ))

            # Add a CharacteristicUserDescriptionDescriptor if description is set and handle is 0
            if (description is not None and handle == 0):
                self.add_descriptor(UserDescription(
                    description=description,
                    characteristic=self
                ))

            # Add additional descriptors
            for desc in descriptors:
                if isinstance(desc, Descriptor):
                    # Bind descriptor to this characteristic
                    desc.characteristic = self

                    # Don't add a CCC descriptor if characteristic already has one.
                    if isinstance(desc, ClientCharacteristicConfig) and self.get_descriptor(ClientCharacteristicConfig):
                        continue

                    # Add descriptor to our list of descriptors
                    self.add_descriptor(desc)

    @classmethod
    def _build(cls, instance):
        """Build a new characteristic object based on this model."""
        # Create a new list of descriptors
        descriptors = []
        for desc in instance.descriptors():
            descriptors.append(desc.build())

        # Create a new object based on current template.
        charac = cls(
            instance.uuid,
            handle=0,
            end_handle=0,
            value=instance.value,
            properties=instance.properties,
            security=instance.security,
            descriptors=descriptors
        )
        if instance.alias is not None:
            charac.alias = instance.alias
        return charac

    def build(self):
        """Build a new characteristic from current object."""
        return self.__class__._build(self)

    def payload(self):
        """Return characteristic payload
        """
        return pack('<BH', self.__properties, self.__value_handle) + self.__value.uuid.packed

    def set_notification_callback(self, callback):
        """Save the provided callback as notification callback
        """
        self.__notification_callback = callback

    def set_indication_callback(self, callback):
        """Save the provided callback as indication callback
        """
        self.__indication_callback = callback

    @Attribute.handle.setter
    def handle(self, new_handle):
        """Set new handle value

        Update the characteristic's value handle as well as handles
        of all descriptors attached to this characteristic.
        """
        if isinstance(new_handle, int):

            # Set attribute handle
            Attribute.handle.fset(self, new_handle)

            self.__value.handle = new_handle + 1
            self.__value_handle = new_handle + 1

            handle = self.__value_handle

            # Update descriptors handle
            for descriptor in self.__descriptors:
                handle += 1
                descriptor.handle = handle

            # Update end handle
            self.__end_handle = handle
        else:
            raise InvalidHandleValueException

    @property
    def value(self):
        return self.__value.value

    @value.setter
    def value(self, value):
        """Set characteristic value

        :param bytes new_value: Value
        """
        self.__value.value = value

        # Notify or indicate if required
        if self.must_notify() and self.__notification_callback is not None:
            self.__notification_callback(self)
        elif self.must_indicate() and self.__indication_callback is not None:
            self.__indication_callback(self)

    @property
    def value_attr(self):
        """Associated value attribute
        """
        return self.__value

    @property
    def value_handle(self):
        """Characteristic value handle
        """
        return self.__value_handle

    @value_handle.setter
    def value_handle(self, value):
        """Update characteristic value handle
        """
        self.__value_handle = value
        self.__value.handle = value

    @property
    def properties(self):
        """Characteristic properties
        """
        return self.__properties

    @properties.setter
    def properties(self, new_properties):
        """Set characteristic properties
        """
        self.__properties = new_properties

    @property
    def uuid(self) -> UUID:
        """Characteristic UUID
        """
        return self.__charac_uuid

    @uuid.setter
    def uuid(self, value):
        """Update characteristic UUID
        """
        self.__charac_uuid = value

    @property
    def end_handle(self):
        """Characteristic end handle
        """
        return self.__end_handle

    @property
    def name(self):
        """Characteristic standard name (if any)
        """
        # Search Bluetooth known UUIDs
        alias = get_uuid_alias(self.__charac_uuid)
        if alias is not None:
            return f"{alias} (0x{self.__charac_uuid})"

        # Search in collaborative CLUES database
        alias = CluesDb.get_uuid_alias(self.__charac_uuid)
        if alias is not None:
            if self.__charac_uuid.type == UUID.TYPE_16:
                return f"{alias} (0x{self.__charac_uuid})"
            else:
                return f"{alias} ({self.__charac_uuid})"


        return str(self.__charac_uuid)

    @property
    def service(self):
        """Service this characteristic belongs to."""
        return self.__service

    @service.setter
    def service(self, service):
        """Set characteristic service."""
        self.__service = service

    @property
    def alias(self) -> Optional[str]:
        """Characteristic alias."""
        return self.__alias

    @alias.setter
    def alias(self, alias: str):
        self.__alias = alias

    ##########################
    # Methods
    ##########################

    def attach(self, service):
        """Attach this characteristic to a service.

        :param service: Service referenced
        :type  service: Service
        """
        self.__service = service

    def get_required_handles(self) -> int:
        if self.handle is not None and self.__end_handle is not None:
            return (self.__end_handle - self.handle) + 1
        return 0

    def readable(self):
        """Determine if characteristic can be read
        """
        return (self.properties & Properties.READ) != 0

    def writeable(self):
        """Determine if characteristic can be written to
        """
        return (
            ((self.properties & Properties.WRITE) != 0) or
            ((self.properties & Properties.WRITE_WITHOUT_RESPONSE) != 0)
        )

    def can_notify(self) -> bool:
        """Determine if characteristic sends notifications.

        :return: ``True`` if characteristic sends notification, ``False`` otherwise.
        :rtype: bool
        """
        return (self.properties & Properties.NOTIFY) != 0

    def must_notify(self):
        """Determine if a notification must be sent for this characteristic.

        Notification must be sent when a characteristic has the notification property and
        its ClientCharacteristicConfiguration descriptor has notifications enabled.
        """
        if (self.properties & Properties.NOTIFY) != 0:
            cccd = self.get_client_config()
            if cccd is not None:
                return cccd.config == 0x0001
        return False

    def can_indicate(self) -> bool:
        """Determine if characteristic sends indications.

        :return: ``True`` if characteristic sends indication, ``False`` otherwise.
        :rtype: bool
        """
        return (self.properties & Properties.INDICATE) != 0

    def must_indicate(self):
        """Determine if an indication must be sent for this characteristic.

        Indication must be sent when a characteristic has the indication property and
        its ClientCharacteristicConfiguration descriptor has indications enabled.
        """
        if (self.properties & Properties.INDICATE) != 0:
            cccd = self.get_client_config()
            if cccd is not None:
                return cccd.config == 0x0002
        return False

    def add_descriptor(self, descriptor: Descriptor) -> 'Characteristic':
        """Add a descriptor

        :param descriptor: Descriptor instance to add to this characteristic.
        :type  descriptor: :class:`whad.ble.profile.characteristic.Descriptor`
        """
        # Set descriptor's handle and update end handle
        if descriptor.handle == 0:
            descriptor.handle = self.__end_handle + 1

        # Add this descriptor to our list of descriptors
        self.__descriptors.append(descriptor)

        # Update our end handle
        self.__end_handle = max(descriptor.handle, self.__end_handle)

        return self

    def get_descriptor(self, desc_type: Union[UUID, Type[Descriptor]]) -> Optional[Descriptor]:
        """Retrieve a decriptor based on its type UUID or class.

        :param desc_type: Descriptor type
        :type  desc_type: UUID, :class:`whad.ble.profile.characteristic.Descriptor`
        :return: First matching descriptor belonging to this characteristic
        :rtype: :class:`whad.ble.profile.Descriptor`, optional
        """
        # Validate descriptor type (UUID or class)
        if isinstance(desc_type, UUID):
            # Descriptor's type UUID is provided, use it as-is
            type_uuid = desc_type
        elif issubclass(desc_type, Descriptor):
            # Descriptor class provided, search for corresponding type UUID
            for desc in self.__descriptors:
                if isinstance(desc, desc_type):
                    return desc
            return None

        # If we found a valid type UUID, look for a matching descriptor
        for desc in self.__descriptors:
            if desc.type_uuid == type_uuid:
                return desc

        # Not found
        return None

    def descriptors(self) -> Iterator[Descriptor]:
        """Iterate over the registered descriptors (generator)
        """
        yield from self.__descriptors

    def get_client_config(self):
        """Find characteristic client configuration descriptor
        """
        for desc in self.__descriptors:
            if isinstance(desc, ClientCharacteristicConfig):
                return desc
        return None

    @property
    def security(self) -> SecurityAccess:
        """Returns security access property
        """
        return self.__security

    def get_security_access(self, access_type):
        """Returns the security access properties linked to an access type.

        :param access_type: access type to check
        :type access_type: SecurityAccess
        """
        for access in self.__security:
            if isinstance(access, access_type):
                return access
        return None

    def check_security_property(self, access_type, prop):
        """Returns a boolean indicating if a property is required for a given access type.

        :param access_type: access type to check
        :type access_type: SecurityAccess
        :param property: security property to check
        :type property: SecurityProperty
        """
        access = self.get_security_access(access_type)
        if access is not None:
            return prop in access.access
        return False



class HIDLocaleNotFound(Exception):
    def __init__(self, locale):
        super().__init__()
        self.locale = locale

    def __str__(self):
        return 'HIDLocaleNotFound(%s)' % self.locale

    def __repr__(self):
        return str(self)


class HIDKeyNotFound(Exception):
    def __init__(self, key, ctrl, alt, shift, gui):
        self.key = key
        self.ctrl = ctrl
        self.alt = alt
        self.shift = shift
        self.gui = gui
        super().__init__()

    def __str__(self):
        return 'HIDKeyNotFound(key={}, ctrl={}, alt={}, shift={}, gui={})'.format(
            self.key,
            self.ctrl,
            self.alt,
            self.shift,
            self.gui
        )

    def __repr__(self):
        return str(self)


class HIDCodeNotFound(Exception):
    def __init__(self, hid_code, modifiers):
        self.modifiers = modifiers
        self.hid_code = hid_code
        super().__init__(self)


    def __str__(self):
        return 'HIDCodeNotFound(hid_code={}, modifiers={}))'.format(
            self.hid_code,
            self.modifiers
        )

    def __repr__(self):
        return str(self)



"""
This module provides some constants used by WHAD to communicate with the Ubertooth One.
"""
from enum import IntEnum

class UbertoothInternalState(IntEnum):
    """Internal states.
    """
    NONE = 0
    ADVERTISEMENT_SNIFFING = 1
    NEW_CONNECTION_SNIFFING = 2
    EXISTING_CONNECTION_SNIFFING = 3
    ACCESS_ADDRESS_SNIFFING = 4

class UbertoothId(IntEnum):
    """USB vendor and product IDs.
    """
    UBERTOOTH_ID_VENDOR         = 0x1D50
    UBERTOOTH_ID_PRODUCT        = 0x6002

class UbertoothTransfers(IntEnum):
    """Input/Output constants.
    """
    CTRL_IN                     = 0xC0
    CTRL_OUT                    = 0x40


class UbertoothModulations(IntEnum):
    """Supported modulations.
    """
    MOD_BT_BASIC_RATE           = 0
    MOD_BT_LOW_ENERGY           = 1
    MOD_80211_FHSS              = 2

class UbertoothCommands(IntEnum):
    """Ubertooth supported commands.
    """
    UBERTOOTH_PING              = 0
    UBERTOOTH_RX_SYMBOLS        = 1
    UBERTOOTH_TX_SYMBOLS        = 2
    UBERTOOTH_GET_USRLED        = 3
    UBERTOOTH_SET_USRLED        = 4
    UBERTOOTH_GET_RXLED         = 5
    UBERTOOTH_SET_RXLED         = 6
    UBERTOOTH_GET_TXLED         = 7
    UBERTOOTH_SET_TXLED         = 8
    UBERTOOTH_GET_1V8           = 9
    UBERTOOTH_SET_1V8           = 10
    UBERTOOTH_GET_CHANNEL       = 11
    UBERTOOTH_SET_CHANNEL       = 12
    UBERTOOTH_RESET             = 13
    UBERTOOTH_GET_SERIAL        = 14
    UBERTOOTH_GET_PARTNUM       = 15
    UBERTOOTH_GET_PAEN          = 16
    UBERTOOTH_SET_PAEN          = 17
    UBERTOOTH_GET_HGM           = 18
    UBERTOOTH_SET_HGM           = 19
    UBERTOOTH_TX_TEST           = 20
    UBERTOOTH_STOP              = 21
    UBERTOOTH_GET_MOD           = 22
    UBERTOOTH_SET_MOD           = 23
    UBERTOOTH_SET_ISP           = 24
    UBERTOOTH_FLASH             = 25
    BOOTLOADER_FLASH            = 26
    UBERTOOTH_SPECAN            = 27
    UBERTOOTH_GET_PALEVEL       = 28
    UBERTOOTH_SET_PALEVEL       = 29
    UBERTOOTH_REPEATER          = 30
    UBERTOOTH_RANGE_TEST        = 31
    UBERTOOTH_RANGE_CHECK       = 32
    UBERTOOTH_GET_REV_NUM       = 33
    UBERTOOTH_LED_SPECAN        = 34
    UBERTOOTH_GET_BOARD_ID      = 35
    UBERTOOTH_SET_SQUELCH       = 36
    UBERTOOTH_GET_SQUELCH       = 37
    UBERTOOTH_SET_BDADDR        = 38
    UBERTOOTH_START_HOPPING     = 39
    UBERTOOTH_SET_CLOCK         = 40
    UBERTOOTH_GET_CLOCK         = 41
    UBERTOOTH_BTLE_SNIFFING     = 42
    UBERTOOTH_GET_ACCESS_ADDRESS= 43
    UBERTOOTH_SET_ACCESS_ADDRESS= 44
    UBERTOOTH_DO_SOMETHING      = 45
    UBERTOOTH_DO_SOMETHING_REPLY= 46
    UBERTOOTH_GET_CRC_VERIFY    = 47
    UBERTOOTH_SET_CRC_VERIFY    = 48
    UBERTOOTH_POLL              = 49
    UBERTOOTH_BTLE_PROMISC      = 50
    UBERTOOTH_SET_AFHMAP        = 51
    UBERTOOTH_CLEAR_AFHMAP      = 52
    UBERTOOTH_READ_REGISTER     = 53
    UBERTOOTH_BTLE_SLAVE        = 54
    UBERTOOTH_GET_COMPILE_INFO  = 55
    UBERTOOTH_BTLE_SET_TARGET   = 56
    UBERTOOTH_BTLE_PHY          = 57
    UBERTOOTH_WRITE_REGISTER    = 58
    UBERTOOTH_JAM_MODE          = 59
    UBERTOOTH_EGO               = 60
    UBERTOOTH_AFH               = 61
    UBERTOOTH_HOP               = 62
    UBERTOOTH_TRIM_CLOCK        = 63
    UBERTOOTH_WRITE_REGISTERS   = 65
    UBERTOOTH_READ_ALL_REGISTERS= 66
    UBERTOOTH_RX_GENERIC        = 67
    UBERTOOTH_TX_GENERIC_PACKET = 68
    UBERTOOTH_FIX_CLOCK_DRIFT   = 69
    UBERTOOTH_CANCEL_FOLLOW     = 70
    UBERTOOTH_LE_SET_ADV_DATA   = 71

class UbertoothModes(IntEnum):
    """Ubertooth modes.
    """
    MODE_IDLE                   = 0
    MODE_RX_SYMBOLS             = 1
    MODE_TX_SYMBOLS             = 2
    MODE_TX_TEST                = 3
    MODE_SPECAN                 = 4
    MODE_RANGE_TEST             = 5
    MODE_REPEATER               = 6
    MODE_LED_SPECAN             = 7
    MODE_BT_FOLLOW              = 8
    MODE_BT_FOLLOW_LE           = 9
    MODE_BT_PROMISC_LE          = 10
    MODE_RESET                  = 11
    MODE_BT_SLAVE_LE            = 12

class UbertoothJammingModes(IntEnum):
    """Jamming modes.
    """
    JAM_NONE                    = 0
    JAM_ONCE                    = 1
    JAM_CONTINUOUS              = 2



"""
Host/Controller interface constants.
"""
from enum import IntEnum

from whad.hub.discovery import Capability
from whad.hub.ble import Commands
from whad.ble.stack.constants import BT_MANUFACTURERS

class HCIInternalState(IntEnum):
    """Host/controller interface state enumeration.
    """
    NONE = 0
    SCANNING = 1
    CENTRAL = 2
    PERIPHERAL = 3

class HCIConnectionState(IntEnum):
    """Host/controller interface connection state
    """
    DISCONNECTED = 0
    INITIATING = 1
    ESTABLISHED = 2

LE_STATES = {
    0 : ("Non-connectable Advertising State", 0 ,
         [Commands.Start, Commands.Stop, Commands.AdvMode, Commands.SetAdvData]
        ),
    1 : ("Scannable Advertising State", 0,
         [Commands.Start, Commands.Stop, Commands.AdvMode, Commands.SetAdvData]
        ),
    2 : ("Connectable Advertising State", Capability.SimulateRole,
         [Commands.Start, Commands.Stop, Commands.AdvMode, Commands.SetAdvData]
        ),
    3 : ("High Duty Cycle Directed Advertising State", 0,
         [Commands.Start, Commands.Stop, Commands.AdvMode, Commands.SetAdvData]
        ),
    4 : ("Passive Scanning State", Capability.Scan,
         [Commands.Start, Commands.Stop, Commands.ScanMode]
        ),
    5 : ("Active Scanning State", Capability.Scan,
         [Commands.Start, Commands.Stop, Commands.ScanMode]
        ),
    6 : ("Initiating State and Connection State (Central Role)", Capability.SimulateRole,
         [
             Commands.Start, Commands.Stop, Commands.CentralMode, Commands.ConnectTo,
             Commands.Disconnect, Commands.SendPDU, Commands.SetEncryption
         ]
        ),
    7 : ("Connection State (Peripheral Role)", Capability.SimulateRole,
         [
            Commands.Start, Commands.Stop, Commands.PeripheralMode, Commands.SendPDU,
            Commands.Disconnect, Commands.SetEncryption
         ]
        )
}

ADDRESS_MODIFICATION_VENDORS = [
    BT_MANUFACTURERS[0].encode("utf-8"),
    BT_MANUFACTURERS[10].encode("utf-8"),
    BT_MANUFACTURERS[13].encode("utf-8"),
    BT_MANUFACTURERS[15].encode("utf-8"),
    BT_MANUFACTURERS[18].encode("utf-8"),
    BT_MANUFACTURERS[48].encode("utf-8"),
    BT_MANUFACTURERS[57].encode("utf-8")
]



from scapy.packet import Packet, bind_layers
from scapy.fields import ByteEnumField, ByteField, LEIntField, LEShortField, \
    XLE3BytesField, ShortField, SignedByteField, XByteField, XLEIntField
from scapy.layers.bluetooth4LE import BTLE
from whad.ble.utils.phy import FieldsSize
from struct import unpack
# Ubertooth Packet Types
UBERTOOTH_PACKET_TYPES = [
    "BR_PACKET",
    "LE_PACKET",
    "MESSAGE",
    "KEEP_ALIVE",
    "SPECAN",
    "LE_PROMISC",
    "EGO_PACKET"
]

# Scapy packets definitions
class Ubertooth_Hdr(Packet):
    name = "Ubertooth header"
    fields_desc = [
        ByteEnumField("packet_type", 0, UBERTOOTH_PACKET_TYPES),
        ByteField("status", None),
        ByteField("channel", None),
        ByteField("clkn_high",None),
        LEIntField("clk_100ns", None),
        SignedByteField("rssi_max", None),
        SignedByteField("rssi_min", None),
        SignedByteField("rssi_avg", None),
        ByteField("rssi_count", None),
        ShortField("unused",None)
    ]
    def pre_dissect(self,s):
        if s[0] == 0x01:
            size = unpack('B',s[14:][5:6])[0]
            return s[0:14+size+FieldsSize.ACCESS_ADDRESS_SIZE+FieldsSize.HEADER_SIZE+FieldsSize.CRC_SIZE]
        else:
            return s

class BTLE_Promiscuous_Data(Packet):
    name = "BTLE Promiscuous Data"
    fields_desc = [
            ByteEnumField("state",None,{0x00 : "access_address",
                           0x01 : "crc_init",
                           0x02 : "hop_interval",
                           0x03 : "hop_increment"})
            ]

class BTLE_Promiscuous_Access_Address(Packet):
    name = "BTLE Promiscuous Access Address"
    fields_desc = [XLEIntField("access_address",None)]

class BTLE_Promiscuous_CRCInit(Packet):
    name = "BTLE Promiscuous CRCInit"
    fields_desc = [XLE3BytesField("crc_init",None)]

class BTLE_Promiscuous_Hop_Interval(Packet):
    name = "BTLE Promiscuous Hop Interval"
    fields_desc = [LEShortField("hop_interval",None)]

class BTLE_Promiscuous_Hop_Increment(Packet):
    name = "BTLE Promiscuous Hop Increment"
    fields_desc = [XByteField("hop_increment",None)]


bind_layers(Ubertooth_Hdr, BTLE, packet_type=0x01)
bind_layers(Ubertooth_Hdr, BTLE_Promiscuous_Data,packet_type=0x05)
bind_layers(BTLE_Promiscuous_Data, BTLE_Promiscuous_Access_Address, state=0x00)
bind_layers(BTLE_Promiscuous_Data, BTLE_Promiscuous_CRCInit, state=0x01)
bind_layers(BTLE_Promiscuous_Data, BTLE_Promiscuous_Hop_Interval, state=0x02)
bind_layers(BTLE_Promiscuous_Data, BTLE_Promiscuous_Hop_Increment, state=0x03)



from whad.dot15d4.stack.mac.constants import MACAddressMode
from whad.zigbee.stack.aps.constants import APSDestinationAddressMode

from whad.zigbee.stack.apl.cluster import Cluster
from whad.zigbee.stack.apl.zcl.commands import ZCLCommands
from whad.zigbee.stack.apl.zcl.attributes import ZCLAttributes
from whad.zigbee.stack.apl.zcl.exceptions import ZCLCommandNotFound
from whad.zigbee.stack.apl.zcl.configuration import ZCLClusterConfiguration
from whad.zigbee.stack.apl.zcl.constants import ZCLClusterType

from scapy.layers.zigbee import ZigbeeClusterLibrary, ZCLGeneralReadAttributes
from whad.scapy.layers.zll import ZCLGeneralDiscoverAttributes, \
    ZCLGeneralDiscoverAttributesResponse, ZigbeeZLLCommissioningCluster

from inspect import stack,signature
from time import time
import logging

logger = logging.getLogger(__name__)


class ZCLClusterMetaclass(type):
    """
    Metaclass allowing to facilitate definition of Zigbee Cluster Library Clusters.
    """
    def __new__(cls, name, bases, attrs):
        # Define lists for cluster specific & profile wide commands
        cluster_specific_commands = {}
        profile_wide_commands = {}

        # Iterate over every commands callbacks
        for key, val in attrs.items():
            # if method is marked as a receive callback or generation callback
            receive_property = getattr(val, "_command_receive", None)
            generate_property = getattr(val, "_command_generate", None)

            # If it is a reception callback
            if receive_property is not None:
                # Get the command ID & name + if it is a profile wide command
                command_id, command_name, profile_wide = receive_property

                # Populate or update the right dictionnary
                commands = profile_wide_commands if profile_wide else cluster_specific_commands

                if command_id in commands:
                    # Update entry
                    commands[command_id]["name"] = command_name
                    commands[command_id]["receive_callback"] = val
                else:
                    # Create entry
                    commands[command_id] = {
                        "name":command_name,
                        "receive_callback": val,
                        "generate_callback":None
                    }

            # If it is a generation callback
            if generate_property is not None:

                # Get the command ID & name + if it is a profile wide command
                command_id, command_name, profile_wide = generate_property
                commands = profile_wide_commands if profile_wide else cluster_specific_commands

                # Populate or update the right dictionnary
                if command_id in commands:
                    # Update entry
                    commands[command_id]["name"] = command_name
                    commands[command_id]["generate_callback"] = val
                else:
                    # Create entry
                    commands[command_id] = {
                        "name":command_name,
                        "receive_callback": None,
                        "generate_callback": val
                    }

        # Generate new attributes according to the created dictionnaries
        if cluster_specific_commands != {}:
            attrs["CLUSTER_SPECIFIC_COMMANDS"] = cluster_specific_commands
        if profile_wide_commands != {}:
            attrs["PROFILE_WIDE_COMMANDS"] = profile_wide_commands

        # Iterate over attributes and populate the ZCL attributes accordingly
        attrs["ATTRIBUTES"] = {}
        # Split the annotations
        if "__annotations__" in attrs:
            for attribute, properties in attrs["__annotations__"].items():
                 attrs["ATTRIBUTES"][attribute] = {
                    "id":properties[0],
                    "permissions":properties[1],
                    "value":attrs[attribute]
                }
        # Build the class
        return super().__new__(cls, name, bases, attrs)


class ZCLCluster(Cluster, metaclass=ZCLClusterMetaclass):
    """
    Base class representing a Zigbee Cluster Library Cluster.
    """
    # Current transaction counter
    _transaction_counter = 0

    def __init__(
                    self,
                    cluster_id,
                    type,
                    default_configuration=ZCLClusterConfiguration()
    ):
        super().__init__(cluster_id)
        # Type of the cluster (Client or Server)
        self.type = type

        # Databases of attributes and commands
        self.attributes = ZCLAttributes()
        self.cluster_specific_commands = ZCLCommands()
        self.profile_wide_commands = ZCLCommands()

        # Populate the cluster specific commands database according to CLUSTER_SPECIFIC_COMMANDS dictionnary
        for command_id, command in self.CLUSTER_SPECIFIC_COMMANDS.items():
            generate_callback = None
            receive_callback = None

            if command["generate_callback"] is not None:
                generate_callback = getattr(self, command["generate_callback"].__name__)
            if command["receive_callback"] is not None:
                receive_callback = getattr(self, command["receive_callback"].__name__)

            # Add the command in the cluster specific commands database
            self.cluster_specific_commands.add_command(
                command_id,
                command["name"],
                generate_callback=generate_callback,
                receive_callback=receive_callback
            )

        # Populate the profile wide commands database according to PROFILE_WIDE_COMMANDS dictionnary
        for command_id, command in self.PROFILE_WIDE_COMMANDS.items():
            generate_callback = None
            receive_callback = None

            if command["generate_callback"] is not None:
                generate_callback = getattr(self, command["generate_callback"].__name__)
            if command["receive_callback"] is not None:
                receive_callback = getattr(self, command["receive_callback"].__name__)

            # Add the command in the profile wide commands database
            self.profile_wide_commands.add_command(
                command_id,
                command["name"],
                generate_callback=generate_callback,
                receive_callback=receive_callback
            )
        # Populate the attribute database
        for attribute, properties in self.ATTRIBUTES.items():
            self.attributes.add_attribute(
                properties["id"],
                attribute,
                properties["value"],
                properties["permissions"]
            )

        # Keep the default configuration and the active configuration
        self.default_configuration = default_configuration
        self.active_configuration = None

        # Pending responses
        self.pending_responses = {}
        # Cache keeping the last transaction
        self.last_transaction = None

        # List of destination nodes & the associated endpoints
        self.destinations = []


    @property
    def configuration(self):
        """
        Access the configuration.
        """
        if self.active_configuration is None:
            # if an active configuration is registered, use it.
            return self.default_configuration
        else:
            # Otherwise, use the default one.
            configuration = self.active_configuration
            self.active_configuration = None
            return configuration


    def connect(self, destination, endpoint):
        """
        Connect this cluster to a specific destination node endpoint.
        """
        for existing_destination in self.destinations:
            if existing_destination["address"] == destination and existing_destination["endpoint"] == endpoint:
                return

        self.destinations.append(
            {
                "address": destination,
                "endpoint": endpoint
            }
        )


    def disconnect(self, *destinations):
        """
        Disconnect this cluster from destination node endpoint(s).
        """
        removed_indexes = []
        for i in range(len(self.destinations)):
            if self.destinations[i] in destinations:
                removed_indexes.append(i)

        for i in removed_indexes:
            del self.destinations[i]


    def wait_response(self, transaction=None, timeout=1):
        """
        Wait for the response associated to a specific transaction (default: last transaction).
        """
        # if transaction not provided, use the last transaction by default
        if transaction is None:
            transaction = self.last_transaction

        # Clear the entry associated to the provided transaction in pending responses dictionnary
        self.pending_responses[transaction] = None

        start = time()
        while self.pending_responses[transaction] is None and time() - start < timeout:
            pass
        # If we got a response return it, otherwise a timeout occurred and we return None
        if self.pending_responses[transaction] is not None:
            return_value = self.pending_responses[transaction]
        else:
            return_value = None
        del self.pending_responses[transaction]
        return return_value


    def configure(
                    self,
                    destination_address_mode = None,
                    destination_address = None,
                    destination_endpoint = None,
                    transaction = None,
                    alias_address=None,
                    alias_sequence_number=0,
                    radius=30,
                    security_enabled_transmission=False,
                    use_network_key=True,
                    acknowledged_transmission=False,
                    fragmentation_permitted=False,
                    include_extended_nonce=False,
                    disable_default_response=False,
                    interpan=False,
                    asdu_handle=0,
                    source_address_mode=MACAddressMode.EXTENDED,
                    destination_pan_id=0xFFFF
    ):
        """
        Update the active configuration.
        """
        self.active_configuration = ZCLClusterConfiguration(
            destination_address_mode = destination_address_mode,
            destination_address = destination_address,
            destination_endpoint = destination_endpoint,
            transaction = transaction,
            alias_address = alias_address,
            alias_sequence_number = alias_sequence_number,
            radius = radius,
            security_enabled_transmission = security_enabled_transmission,
            use_network_key = use_network_key,
            acknowledged_transmission = acknowledged_transmission,
            fragmentation_permitted = fragmentation_permitted,
            include_extended_nonce = include_extended_nonce,
            disable_default_response = disable_default_response,
            interpan=interpan,
            asdu_handle=asdu_handle,
            source_address_mode=source_address_mode,
            destination_pan_id=destination_pan_id
        )



    def send_command(self, command):
        """
        Send a command.
        """
        csc = self.cluster_specific_commands
        pwc = self.profile_wide_commands
        # Find the command infos linked to the caller method
        found_calling_callback = False

        caller_function = getattr(self, stack()[1][3])
        cluster_specific = False

        try:
            command_id, command_structure = csc.get_command_by_callback(caller_function)
            cluster_specific = True
            found_calling_callback = True
        except ZCLCommandNotFound:
            pass

        if not cluster_specific:
            try:
                command_id, command_structure = pwc.get_command_by_callback(caller_function)
                cluster_specific = False
                found_calling_callback = True
            except ZCLCommandNotFound:
                pass

        # if we did not found the related info, returns
        if not found_calling_callback:
            return False

        logger.info(
            "[zcl] Cluster %s (cluster_id=%s) Transmitting %s command '%s' (command_id=%s)",
            self.__class__.__name__,
            hex(self.cluster_id),
            "cluster specific" if cluster_specific else "profile wide",
            command_structure.name,
            hex(command_id),
        )
        # Get the current configuration
        current_configuration = self.configuration

        # If no transaction counter is provided, use the global one
        if current_configuration.transaction is None:
            transaction = ZCLCluster._transaction_counter
            ZCLCluster._transaction_counter += 1
        else:
            transaction = current_configuration.transaction

        # Register the last transaction
        self.last_transaction = transaction

        # Build a ZigbeeClusterLibrary PDU according to current configuration
        asdu = ZigbeeClusterLibrary(
                zcl_frametype=1 if cluster_specific else 0,
                command_direction=(
                    0 if self.type == ZCLClusterType.CLIENT else 1
                ),
                command_identifier=command_id,
                transaction_sequence=transaction,
                disable_default_response=current_configuration.disable_default_response
        ) / command

        # If we need to use interpan, add ZLL Commissioning Cluster header
        if current_configuration.interpan:
            asdu = ZigbeeZLLCommissioningCluster(
                zcl_frametype=1 if cluster_specific else 0,
                direction=0 if self.type == ZCLClusterType.CLIENT else 1,
                command_identifier=command_id,
                transaction_sequence=transaction,
                disable_default_response=int(current_configuration.disable_default_response)
            ) / command

            # Send InterPAN data
            return self.send_interpan_data(
                asdu,
                asdu_handle=current_configuration.asdu_handle,
                source_address_mode=current_configuration.source_address_mode,
                destination_pan_id=current_configuration.destination_pan_id,
                destination_address=current_configuration.destination_address,
                destination_address_mode=current_configuration.destination_address_mode,
                acknowledged_transmission=current_configuration.acknowledged_transmission

            )
        else:
            # Send PAN data, to the destination address if provided, otherwise to registered destinations
            if current_configuration.destination_address is not None:
                return self.send_data(
                    asdu,
                    current_configuration.destination_address_mode,
                    current_configuration.destination_address,
                    current_configuration.destination_endpoint,
                    alias_address=current_configuration.alias_address,
                    alias_sequence_number=current_configuration.alias_sequence_number,
                    radius=current_configuration.radius,
                    security_enabled_transmission=current_configuration.security_enabled_transmission,
                    use_network_key=current_configuration.use_network_key,
                    acknowledged_transmission=current_configuration.acknowledged_transmission,
                    fragmentation_permitted=current_configuration.fragmentation_permitted,
                    include_extended_nonce=current_configuration.include_extended_nonce
                )
            else:
                for destination in self.destinations:
                    self.send_data(
                        asdu,
                        APSDestinationAddressMode.SHORT_ADDRESS_DST_ENDPOINT_PRESENT,
                        destination["address"],
                        destination["endpoint"],
                        alias_address=current_configuration.alias_address,
                        alias_sequence_number=current_configuration.alias_sequence_number,
                        radius=current_configuration.radius,
                        security_enabled_transmission=current_configuration.security_enabled_transmission,
                        use_network_key=current_configuration.use_network_key,
                        acknowledged_transmission=current_configuration.acknowledged_transmission,
                        fragmentation_permitted=current_configuration.fragmentation_permitted,
                        include_extended_nonce=current_configuration.include_extended_nonce
                    )


    def on_interpan_data(self, asdu,destination_pan_id, destination_address, source_pan_id, source_address, link_quality):
        """
        Triggered when an InterPAN data is received by the cluster.

        Forwards the data to the right reception callback.
        """
        # Get command identifier
        command_identifier = asdu.command_identifier

        # Select the right command database depending on the type of command
        commands = (
            self.cluster_specific_commands if
            asdu.zcl_frametype == 1 else
            self.profile_wide_commands
        )
        # Try to find the corresponding reception callback in the database
        try:
            command = commands.get_command_by_id(command_identifier)

            # Raise an error if no command is found
            if command.receive_callback is None:
                raise ZCLCommandNotFound

            # Adapt informations to callbacks parameters, based on signature
            parameters = []
            callback_parameters = signature(command.receive_callback).parameters
            if ZigbeeClusterLibrary in asdu:
                base_class = ZigbeeClusterLibrary
            elif ZigbeeZLLCommissioningCluster in asdu:
                base_class = ZigbeeZLLCommissioningCluster
            for name, parameter in callback_parameters.items():
                if int(parameter.kind) == 1:
                    if name in ("payload","command"):
                        parameters += [asdu[base_class].payload]
                    elif name in ("source","src", "source_address"):
                        parameters += [source_address]
                    elif name in ("source_pan_id", "src_panid","source_panid"):
                        parameters += [source_pan_id]
                    elif name in ("destination","dest", "dest_address","destination_address"):
                        parameters += [destination_address]
                    elif name in ("destination_pan_id","dest_panid", "destination_panid"):
                        parameters += [destination_pan_id]
                    elif name in ("link_quality", "lqi"):
                        parameters += [link_quality]
                    elif name in ("transaction", "transaction_sequence"):
                        parameters += [asdu[base_class].transaction_sequence]
                    elif name in ("no_response", "disable_default_response"):
                        parameters += [asdu[base_class].disable_default_response]

            logger.info(
                "[zcl] Cluster %s (cluster_id=%s) Receiving %s interpan command '%s' (command_id=%s)",
                self.__class__.__name__,
                hex(self.cluster_id),
                "cluster specific" if asdu.zcl_frametype == 1 else "profile wide",
                command.name,
                hex(command_identifier),
            )
            # Call reception callback and update the pending response dict with the return value
            return_value = command.receive_callback(*parameters)
            if asdu.transaction_sequence in self.pending_responses:
                self.pending_responses[asdu.transaction_sequence] = return_value

        except ZCLCommandNotFound:
            logger.info("[zcl] command not found (command_identifier = 0x%02x)", command_identifier)


    def on_data(self, asdu, source_address, source_address_mode, security_status, link_quality):
        """
        Triggered when a data is received by the cluster.

        Forwards the data to the right reception callback.
        """
        command_identifier = asdu.command_identifier
        commands = (
            self.cluster_specific_commands if
            asdu.zcl_frametype == 1 else
            self.profile_wide_commands
        )
        # Try to find the corresponding reception callback in the database
        try:
            command = commands.get_command_by_id(command_identifier)

            # If no reception callback is found, raise an error
            if command.receive_callback is None:
                raise ZCLCommandNotFound
            # Adapt informations to callbacks parameters, based on signature
            parameters = []
            callback_parameters = signature(command.receive_callback).parameters
            for name, parameter in callback_parameters.items():
                if int(parameter.kind) == 1:
                    if name in ("payload","command"):
                        parameters += [asdu[ZigbeeClusterLibrary].payload]
                    elif name in ("source","src", "source_address"):
                        parameters += [source_address]
                    elif name in ("source_mode","source_address_mode","mode"):
                        parameters += [source_address_mode]
                    elif name in ("security", "security_status"):
                        parameters += [security_status]
                    elif name in ("link_quality", "lqi"):
                        parameters += [link_quality]
                    elif name in ("transaction", "transaction_sequence"):
                        parameters += [asdu[ZigbeeClusterLibrary].transaction_sequence]
                    elif name in ("no_response", "disable_default_response"):
                        parameters += [asdu[ZigbeeClusterLibrary].disable_default_response]

            logger.info(
                "[zcl] Cluster %s (cluster_id=%s) Receiving %s command '%s' (command_id=%s)",
                self.__class__.__name__,
                hex(self.cluster_id),
                "cluster specific" if asdu.zcl_frametype == 1 else "profile wide",
                command.name,
                hex(command_identifier),
            )
            # Call reception callback and update the pending response dict with the return value
            return_value = command.receive_callback(*parameters)
            if asdu.transaction_sequence in self.pending_responses:
                self.pending_responses[asdu.transaction_sequence] = return_value

        except ZCLCommandNotFound:
            logger.info("[zcl] command not found (command_identifier = 0x%02x)", command_identifier)


    # Decorators
    def command_receive(command_id, command_name, profile_wide=False):
        """
        Mark a method as command reception callback.
        """
        def receive_decorator(f):
            f._command_receive = (command_id, command_name, profile_wide)
            return f
        return receive_decorator

    def command_generate(command_id, command_name, profile_wide=False):
        """
        Mark a method as command generator callback.
        """
        def generate_decorator(f):
            f._command_generate = (command_id, command_name, profile_wide)
            return f
        return generate_decorator

    # Profile wide commands
    @command_generate(0x00, "Read Attributes", profile_wide=True)
    def read_attributes(self, *attributes):
        """
        Read a list of attributes.
        """
        command = ZCLGeneralReadAttributes(attribute_identifiers=list(attributes))
        self.send_command(command)
        attributes = self.wait_response()
        return attributes

    @command_receive(0x0b, "Default Response", profile_wide=True)
    def on_default_response(self, command):
        """
        Processes a received default response.
        """
        return command.status

    @command_receive(0x01, "Read Attributes Response", profile_wide=True)
    def on_read_attributes_response(self, command):
        """
        Processes a Read Attributes response.
        """
        attributes = []
        for attribute in command.read_attribute_status_record:
            attributes.append(
                (
                    attribute.attribute_identifier,
                    attribute.status,
                    attribute.attribute_value if attribute.status == 0 else None
                )
            )
        return attributes

    @command_generate(0x0c, "Discover Attributes", profile_wide=True)
    def discover_attributes(self, start_identifier=0, max_reports=0xFFFF):
        """
        Discover a set of Attributes.
        """
        attributes = []
        command = ZCLGeneralDiscoverAttributes(
            start_attribute_identifier=start_identifier,
            max_attribute_identifiers=max_reports
        )
        self.send_command(command)
        return self.wait_response()

    @command_receive(0x0d, "Discover Attributes Response", profile_wide=True)
    def on_discover_attributes(self, command):
        """
        Processes a Discover Attributes response.
        """
        attributes = []
        for attribute in command.attribute_records:
            attributes.append(
                (
                    attribute.attribute_identifier,
                    attribute.attribute_data_type
                )
            )
        return (0 == command.discovery_complete, attributes)


    # Black magic to keep attributes database consistent with local variable
    def __setattr__(self, att, value):
        if "attributes" in self.__dict__:
            for attribute in self.__dict__["attributes"].attributes.values():
                if attribute.name == att:
                    attribute.value = value
        return super().__setattr__(att, value)

    def __getattribute__(self, att):
        if att != "__dict__" and "attributes" in self.__dict__:
            for attribute in self.__dict__["attributes"].attributes.values():
                if attribute.name == att:
                    super().__setattr__(att, attribute.value)
        return super().__getattribute__(att)


class ZCLClientCluster(ZCLCluster):
    """
    Base class for Zigbee Cluster Library Client Clusters.
    """
    @classmethod
    def child_clusters(cls):
        """
        Returns a list of child clusters classes.
        """
        subclasses = set()
        work = [cls]
        while work:
            parent = work.pop()
            for child in parent.__subclasses__():
                if child not in subclasses:
                    subclasses.add(child)
                    work.append(child)
        return subclasses

    def __init__(
                    self,
                    cluster_id,
                    default_configuration=ZCLClusterConfiguration()
    ):
        super().__init__(
            cluster_id,
            ZCLClusterType.CLIENT,
            default_configuration=default_configuration
        )

class ZCLServerCluster(ZCLCluster):
    """
    Base class for Zigbee Cluster Library Server Clusters.
    """
    @classmethod
    def child_clusters(cls):
        """
        Returns a list of child clusters classes.
        """
        subclasses = set()
        work = [cls]
        while work:
            parent = work.pop()
            for child in parent.__subclasses__():
                if child not in subclasses:
                    subclasses.add(child)
                    work.append(child)
        return subclasses


    def __init__(
                    self,
                    cluster_id,
                    default_configuration=ZCLClusterConfiguration()
    ):
        super().__init__(
            cluster_id,
            ZCLClusterType.SERVER,
            default_configuration=default_configuration
        )
